# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2015-2020 Rapptz
Copyright (c) 2015 Rossen Georgiev <rossen@rgp.io>
Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

This is a modified version of
https://github.com/Rapptz/discord.py/blob/master/discord/gateway.py
and https://github.com/ValvePython/steam/blob/master/steam/core/cm.py
"""

import asyncio
import concurrent.futures
import functools
import logging
import random
import struct
import sys
import threading
import time
import traceback
from gzip import GzipFile
from io import BytesIO
from typing import TYPE_CHECKING, Callable, Dict, List, NamedTuple, Optional, Tuple, Union

import aiohttp

from . import utils
from .enums import EPersonaState, EResult
from .errors import NoCMsFound
from .iterators import AsyncIterator
from .protobufs import EMsg, Msg, MsgProto

if TYPE_CHECKING:
    from .client import Client
    from .enums import EUIMode
    from .protobufs.steammessages_base import CMsgMulti
    from .protobufs.steammessages_clientserver_login import CMsgClientLogonResponse
    from .state import ConnectionState


__all__ = (
    "SteamWebSocket",
    "ConnectionClosed",
    "WebSocketClosure",
)

log = logging.getLogger(__name__)
EXECUTOR = concurrent.futures.ThreadPoolExecutor()


def return_true(*_):
    return True


class EventListener(NamedTuple):
    emsg: EMsg
    predicate: Callable[..., bool]
    future: asyncio.Future


class ConnectionClosed(Exception):
    def __init__(self, cm: str, cms: "CMServerList"):
        self.cm = cm
        self.cm_list = cms
        super().__init__(f"Connection to {self.cm}, has closed.")


class WebSocketClosure(Exception):
    """An exception to make up for the fact that aiohttp doesn't signal closure."""


class CMServerList(AsyncIterator):
    GOOD = 1
    BAD = 2
    __slots__ = ("dict", "cell_id", "best_cms", "_state")

    def __init__(self, state: "ConnectionState", first_cm_to_try: str):
        super().__init__(state, None, None, None)
        self.dict: Dict[str, Tuple[int, float]] = dict()
        self.best_cms: List[Tuple[str, float]] = []
        self.cell_id = 0
        if first_cm_to_try is not None:
            self.queue.put_nowait(first_cm_to_try)

    def __len__(self):
        return len(self.dict)

    async def fill(self) -> None:
        if not await self.fetch_servers_from_api():
            raise NoCMsFound("No Community Managers could be found to connect to")

        good_servers = [k for (k, v) in self.dict.items() if v[0] == self.GOOD]

        if len(good_servers) == 0:
            log.debug("No good servers left. Resetting...")
            self.reset_all()
            return

        random.shuffle(good_servers)
        await self.ping_cms(good_servers)
        for server_address, _ in self.best_cms:
            self.queue.put_nowait(server_address)

    def clear(self) -> None:
        if self.dict:
            log.debug("List cleared.")
        self.dict.clear()

    async def fetch_servers_from_api(self, cell_id: int = 0) -> bool:
        log.debug("Attempting to fetch servers from the WebAPI")
        self.cell_id = cell_id
        try:
            resp = await self._state.http.get_cm_list(cell_id)
        except Exception as e:
            log.error(f"WebAPI fetch request failed with result: {repr(e)}")
            return False

        resp = resp["response"]
        if resp["result"] != EResult.OK:
            log.error(
                f'Fetching the CMList failed with Result: {EResult(resp["result"])} Message: {repr(resp["message"])}'
            )
            return False

        websockets_list = resp["serverlist_websockets"]
        log.debug(f"Received {len(websockets_list)} servers from WebAPI")

        self.clear()
        self.cell_id = cell_id
        self.merge_list(websockets_list)

        return True

    def reset_all(self) -> None:
        log.debug("Marking all CM servers as Good.")
        for server in self.dict:
            self.mark_good(server)
        self.best_cms = []

    def mark_good(self, server: str, score: float = 0.0) -> None:
        self.dict[server] = self.GOOD, score

    def mark_bad(self, server: str) -> None:
        self.dict[server] = self.BAD, 0.0

    def merge_list(self, hosts: List[str]) -> None:
        total = len(self)
        for host in hosts:
            if host not in self.dict:
                self.mark_good(host)
        if len(self) > total:
            log.debug(f"Added {len(self) - total} new CM server addresses.")

    async def ping_cms(self, hosts: List[str] = None, to_ping: int = 10) -> None:
        hosts = list(self.dict.keys()) if hosts is None else hosts
        for host in hosts[:to_ping]:  # only ping the first 10 cms (by default)
            # TODO dynamically make sure we get good ones
            # by checking len and stuff
            start = time.perf_counter()
            try:
                resp = await self._state.http._session.get(f"https://{host}/cmping/", timeout=5)
                if resp.status != 200:
                    self.mark_bad(host)
                load = resp.headers["X-Steam-CMLoad"]
            except (KeyError, asyncio.TimeoutError):
                self.mark_bad(host)
            else:
                latency = time.perf_counter() - start
                score = (int(load) * 2) + latency
                self.best_cms.append((host, score))
                self.best_cms = sorted(self.best_cms, key=lambda x: x[1])
                self.mark_good(host, score)

        log.debug("Finished pinging CMs")


class KeepAliveHandler(threading.Thread):  # ping commands are cool
    __slots__ = (
        "ws",
        "interval",
        "heartbeat",
        "heartbeat_timeout",
        "msg",
        "block_msg",
        "behind_msg",
        "latency",
        "_stop_ev",
        "_last_ack",
        "_last_send",
        "_main_thread_id",
    )

    def __init__(self, **kwargs):
        super().__init__()
        self.ws: "SteamWebSocket" = kwargs.pop("ws")
        self.interval: int = kwargs.pop("interval")
        self._main_thread_id = self.ws.thread_id
        self.heartbeat = MsgProto(EMsg.ClientHeartBeat)
        self.heartbeat_timeout = 60
        self.msg = "Keeping websocket alive with heartbeat {}."
        self.block_msg = "Heartbeat blocked for more than {} seconds."
        self.behind_msg = "Can't keep up, websocket is {:.1f} behind."
        self._stop_ev = threading.Event()
        self._last_ack = time.perf_counter()
        self._last_send = time.perf_counter()
        self.latency = float("inf")

    def run(self) -> None:
        while not self._stop_ev.wait(self.interval):
            if self._last_ack + self.heartbeat_timeout < time.perf_counter():
                log.warning(f"Server {self.ws.cm} has stopped responding to the gateway. Closing and restarting.")
                coro = self.ws.handle_close()
                f = asyncio.run_coroutine_threadsafe(coro, loop=self.ws.loop)

                try:
                    f.result()
                except Exception:
                    pass
                finally:
                    return self.stop()

            log.debug(self.msg.format(self.heartbeat))
            coro = self.ws.send_as_proto(self.heartbeat)
            f = asyncio.run_coroutine_threadsafe(coro, loop=self.ws.loop)
            try:
                # block until sending is complete
                total = 0
                while 1:
                    try:
                        f.result(10)
                        break
                    except concurrent.futures.TimeoutError:
                        total += 10
                        try:
                            frame = sys._current_frames()[self._main_thread_id]
                        except KeyError:
                            msg = self.block_msg
                        else:
                            stack = traceback.format_stack(frame)
                            msg = f'{self.block_msg}\nLoop thread traceback (most recent call last):\n{"".join(stack)}'
                        log.warning(msg.format(total))

            except Exception:
                self.stop()
            else:
                self.ack()
                self._last_send = time.perf_counter()

    def stop(self) -> None:
        self._stop_ev.set()

    def ack(self) -> None:
        self._last_ack = time.perf_counter()
        self.latency = self._last_ack - self._last_send - self.interval
        if self.latency > 10:
            log.warning(self.behind_msg.format(self.latency))


class SteamWebSocket:
    ___slots___ = (
        "socket ",
        "loop ",
        "_connection ",
        "cm_list ",
        "_dispatch ",
        "cm ",
        "thread_id ",
        "listeners ",
        "connected ",
        "steam_id ",
        "handlers ",
        "_current_job_id ",
        "_parsers ",
    )

    def __init__(
        self, socket: aiohttp.ClientWebSocketResponse, *, loop: asyncio.AbstractEventLoop,
    ):
        self.socket = socket
        self.loop = loop

        self._connection: Optional["ConnectionState"] = None
        self.cm_list: Optional[CMServerList] = None
        self._keep_alive: Optional[KeepAliveHandler] = None
        self._dispatch = lambda *args: None
        self.cm: Optional[str] = None
        self.cell_id = 0
        self.thread_id = threading.get_ident()

        self.listeners: List[EventListener] = []
        self._parsers: Dict[EMsg, Callable[["ConnectionState", "MsgProto"], None]] = dict()

        self.connected = False
        self.session_id = 0
        self.steam_id = 0
        self._current_job_id = 0

        self.handlers = {
            EMsg.Multi: self.handle_multi,
            EMsg.ClientLogOnResponse: self.handle_logon,
        }

    @property
    def latency(self) -> float:
        """:class:`float`: Measures latency between a HEARTBEAT and a HEARTBEAT_ACK in seconds."""
        return self._keep_alive.latency

    def wait_for(self, emsg: EMsg, predicate: Callable[..., bool] = None) -> asyncio.Future:
        future = self.loop.create_future()
        entry = EventListener(emsg=emsg, predicate=predicate or return_true, future=future)
        self.listeners.append(entry)
        return future

    @classmethod
    async def from_client(cls, client: "Client", cm: str = None, cms: CMServerList = None) -> "SteamWebSocket":
        connection = client._connection
        cm_list = cms or CMServerList(connection, cm)
        async for cm in cm_list:
            log.info(f"Creating a websocket connection to: {cm}")
            socket = await client.http.connect_to_cm(cm)
            log.debug(f"Connected to {cm}")
            payload = MsgProto(
                EMsg.ClientLogon,
                account_name=client.username,
                web_logon_nonce=client.token,
                client_os_type=4294966596,
                protocol_version=65580,
                chat_mode=2,
                ui_mode=4,
                qos_level=2,
            )
            ws = cls(socket, loop=client.loop)
            # dynamically add attributes needed
            ws._connection = connection
            ws._parsers = connection.parsers
            ws._dispatch = client.dispatch
            ws.steam_id = client.user.id64
            ws.cm = cm
            ws.cm_list = cm_list
            ws.connected = True
            await ws.send_as_proto(payload)  # send the identification message straight away
            ws._dispatch("connect")
            return ws

    async def poll_event(self) -> None:
        try:
            message = await self.socket.receive()
            if message.type is aiohttp.WSMsgType.ERROR:
                log.debug(f"Received {message}")
                raise message.data
            if message.type in (aiohttp.WSMsgType.CLOSED, aiohttp.WSMsgType.CLOSE):
                log.debug(f"Received {message}")
                raise WebSocketClosure

            message = message.data
            if not message:  # it can sometimes be None/empty
                return
            await self.receive(message)
        except WebSocketClosure:
            await self.handle_close()

    async def receive(self, message: bytes) -> None:
        self._dispatch("socket_raw_receive", message)

        emsg_value = struct.unpack_from("<I", message)[0]
        emsg = EMsg(utils.clear_proto_bit(emsg_value))
        if emsg in self.handlers:
            msg = MsgProto(emsg, message)
            self._dispatch("socket_receive", msg)
            return await self.handlers[emsg](msg)

        if not self.connected:
            return log.debug(f"Dropped unexpected message: {repr(emsg)} {repr(message)}")
        try:
            if utils.is_proto(emsg_value):
                msg = MsgProto(emsg, message)
            else:
                msg = Msg(emsg, message, extended=True)
            log.debug(f"Socket has received {repr(msg)} from the websocket.")
        except Exception as e:
            log.fatal(f"Failed to deserialize message: {repr(emsg)}, {repr(message)}")
            if emsg != EMsg.ServiceMethodResponse:  # the repr likely failed so just ignore it
                return log.exception(e)

        self._dispatch("socket_receive", msg)

        try:
            func = self._parsers[emsg]
        except KeyError:
            log.debug(f"Ignoring event {repr(msg)}")
        else:
            await utils.maybe_coroutine(func, self._connection, msg)

        # remove the dispatched listener
        removed = []
        for index, entry in enumerate(self.listeners):
            if entry.emsg != emsg:
                continue

            future = entry.future
            if future.cancelled():
                removed.append(index)
                continue

            try:
                valid = entry.predicate(msg)
            except Exception as exc:
                future.set_exception(exc)
                removed.append(index)
            else:
                if valid:
                    future.set_result(msg)
                    removed.append(index)

        for index in reversed(removed):
            del self.listeners[index]

    async def send(self, data: bytes) -> None:
        self._dispatch("socket_raw_send", data)
        await self.socket.send_bytes(data=data)

    async def send_as_proto(self, message: Union[MsgProto, Msg]) -> None:
        message.steam_id = self.steam_id
        message.session_id = self.session_id

        self._dispatch("socket_send", message)
        await self.send(bytes(message))

    async def close(self, code: int = 4000) -> None:
        if self._keep_alive is not None:
            self._keep_alive.stop()
        if self.connected:
            await self.send_as_proto(MsgProto(EMsg.ClientLogOff))
        await self.socket.close(code=code)

    async def handle_close(self):
        try:
            await self.close()
        except Exception:
            pass
        self.cm_list.queue.get_nowait()  # pop the disconnected cm
        log.info(f"Websocket closed, cannot reconnect.")
        raise ConnectionClosed(self.cm, self.cm_list)

    async def handle_logon(self, msg: MsgProto["CMsgClientLogonResponse"]) -> None:
        if msg.body.eresult == EResult.OK:
            log.debug("Logon completed")

            self.session_id = msg.session_id
            self.cell_id = self.cm_list.cell_id = msg.body.cell_id

            interval = msg.body.out_of_game_heartbeat_seconds
            self._keep_alive = KeepAliveHandler(ws=self, interval=interval)
            self._keep_alive.start()
            log.debug("Heartbeat started.")

            await self.send_um("ChatRoom.GetMyChatRoomGroups#1_Request")
            status = MsgProto(EMsg.ClientChangeStatus, persona_state=EPersonaState.Online)
            await self.send_as_proto(status)
            # setting your status to offline will stop you receiving persona updates, don't ask why.
            await self.send_as_proto(MsgProto(EMsg.ClientRequestCommentNotifications))
        else:
            if msg.body.eresult == EResult.InvalidPassword:
                await asyncio.sleep(60)
            raise ConnectionClosed(self.cm, self.cm_list)

    async def handle_multi(self, msg: MsgProto["CMsgMulti"]) -> None:
        log.debug("Received a multi, unpacking")
        if msg.body.size_unzipped:
            log.debug(f"Decompressing payload ({len(msg.body.message_body)} -> {msg.body.size_unzipped})")
            # aiofiles is overrated :P
            bytes_io = await self.loop.run_in_executor(EXECUTOR, BytesIO, msg.body.message_body)
            gzipped = await self.loop.run_in_executor(EXECUTOR, functools.partial(GzipFile, fileobj=bytes_io))
            data = await self.loop.run_in_executor(EXECUTOR, gzipped.read)
            if len(data) != msg.body.size_unzipped:
                return log.info(f"Unzipped size mismatch for multi payload {msg}, discarding")
        else:
            data = msg.body.message_body

        while len(data) > 0:
            size = struct.unpack_from("<I", data)[0]
            await self.receive(data[4 : 4 + size])
            data = data[4 + size :]

    async def send_um(self, name: str, **kwargs) -> int:
        msg = MsgProto(EMsg.ServiceMethodCallFromClient, um_name=name, **kwargs)
        msg.header.job_id_source = self._current_job_id = (self._current_job_id + 1) % 10000 or 1
        await self.send_as_proto(msg)
        return self._current_job_id

    async def change_presence(
        self, *, games: List[dict], state: EPersonaState, ui_mode: "EUIMode", force_kick: bool,
    ) -> None:
        if force_kick:
            kick = MsgProto(EMsg.ClientKickPlayingSession)
            log.debug("Kicking any currently playing sessions")
            await self.send_as_proto(kick)
        if games:
            activity = MsgProto(EMsg.ClientGamesPlayedWithDataBlob, games_played=games)
            log.debug(f"Sending {activity} to change activity")
            await self.send_as_proto(activity)
        if state:
            state = MsgProto(EMsg.ClientPersonaState, status_flags=state)
            log.debug(f"Sending {state} to change state")
            await self.send_as_proto(state)
        if ui_mode:
            ui_mode = MsgProto(EMsg.ClientCurrentUIMode, uimode=ui_mode)
            log.debug(f"Sending {ui_mode} to change UI mode")
            await self.send_as_proto(ui_mode)



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from typing import TYPE_CHECKING

from ... import utils
from ...game import Game
from .errors import BadArgument

if TYPE_CHECKING:
    from steam.ext import commands

    from ...channel import BaseChannel
    from ...clan import Clan
    from ...group import Group
    from ...user import User
    from .context import Context

__all__ = (
    "Converter",
    "UserConverter",
    "ChannelConverter",
    "ClanConverter",
    "GroupConverter",
    "GameConverter",
    "Default",
    "Author",
    "DefaultChannel",
    "DefaultClan",
    "DefaultGroup",
    "DefaultGame",
)


class Converter:
    """A custom class from which converters can be derived.
    They should be type-hinted to a command's argument.

    Some custom types from this library can be type-hinted
    just using their normal type (see below).

    Examples
    --------

    Builtin: ::

        @bot.command()
        async def command(ctx, user: steam.User):
            # this will end up making the user variable a `User` object.

        # invoked as
        # !command 80528701850124288
        # or !command "Gobot1234"

    A custom converter: ::

        class ImageConverter:
            async def convert(self, ctx: 'commands.Context', argument: str):
                async with aiohttp.ClientSession as session:
                    async with session.get(argument) as r:
                        image_bytes = await r.read()
                try:
                    return steam.Image(image_bytes)
                except (TypeError, ValueError) as exc:  # failed to convert to an image
                    raise commands.BadArgument from exc

        # then later

        @bot.command()
        async def set_avatar(ctx, avatar: ImageConverter):
            await bot.edit(avatar=avatar)
        await ctx.send('👌')

        # invoked as
        # !set_avatar "my image url"
    """

    async def convert(self, ctx: "commands.Context", argument: str):
        raise NotImplementedError("Derived classes must implement this")

    def __repr__(self):
        return self.__class__.__name__


class UserConverter(Converter):
    """The converter that is used when the
    type-hint passed is :class:`~steam.User`.

    Lookup is in the order of:
        - steam id
        - name
    """

    async def convert(self, ctx: "commands.Context", argument: str) -> "User":
        user = ctx.bot.get_user(argument) or await ctx.bot.fetch_user(argument)
        if user is None:
            user = utils.get(ctx.bot.users, name=argument)
        if user is None:
            raise BadArgument(f'Failed to convert "{argument}" to a Steam user')
        return user


class ChannelConverter(Converter):
    """The converter that is used when the
    type-hint passed is :class:`~steam.Channel`.

    Lookup is in the order of:
        - id
        - name
    """

    async def convert(self, ctx: "commands.Context", argument: str) -> "BaseChannel":
        if argument.isdigit():
            channel = ctx.bot._connection._combined.get(int(argument))
        else:
            if ctx.clan:
                channel = utils.get(ctx.clan.channels, name=argument)
            elif ctx.group:
                channel = utils.get(ctx.group.channels, name=argument)
            else:
                channel = None
        if channel is None:
            raise BadArgument(f'Failed to convert "{argument}" to a Steam group')
        return channel


class ClanConverter(Converter):
    """The converter that is used when the
    type-hint passed is :class:`~steam.Clan`.

    Lookup is in the order of:
        - steam id
        - name
    """

    async def convert(self, ctx: "commands.Context", argument: str) -> "Clan":
        clan = ctx.bot.get_clan(argument)
        if clan is None:
            clan = utils.get(ctx.bot.clans, name=argument)
        if clan is None:
            raise BadArgument(f'Failed to convert "{argument}" to a Steam clan')
        return clan


class GroupConverter(Converter):
    """The converter that is used when the
    type-hint passed is :class:`~steam.Group`.

    Lookup is in the order of:
        - id
        - name
    """

    async def convert(self, ctx: "commands.Context", argument: str) -> "Group":
        if argument.isdigit():
            group = ctx.bot.get_group(int(argument))
        else:
            group = utils.get(ctx.bot.clans, name=argument)
        if group is None:
            raise BadArgument(f'Failed to convert "{argument}" to a Steam group')
        return group


class GameConverter(Converter):
    """The converter that is used when the
    type-hint passed is :class:`~steam.Game`.

    If the param is a digit it is assumed
    that the argument is the game's app id else
    it is assumed it is the game's title.
    """

    async def convert(self, ctx: "commands.Context", argument: str):
        return Game(app_id=int(argument)) if argument.isdigit() else Game(title=argument)


class Default:
    """A custom way to specify a default values for commands.

    Examples
    --------
    Builtin: ::

        @bot.command()
        async def info(ctx, user: steam.User = Author):
            # if no user is passed it will be ctx.author

    A custom default: ::

        class CurrentCommand(commands.Default):
            async def default(self, ctx: 'commands.Context'):
                return ctx.command  # return the current command

        # then later

        @bot.command()
        async def source(ctx, command=CurrentCommand):
            # command would now be source
    """

    async def default(self, ctx: "commands.Context"):
        raise NotImplementedError("derived classes need to implement this")


class Author(Default):
    """Returns the :attr:`.Context.author`"""

    async def default(self, ctx: "commands.Context"):
        return ctx.author


class DefaultChannel(Default):
    """Returns the :attr:`.Context.channel`"""

    async def default(self, ctx: "commands.Context"):
        return ctx.channel


class DefaultGroup(Default):
    """Returns the :attr:`.Context.group`"""

    async def default(self, ctx: "commands.Context"):
        return ctx.group


class DefaultClan(Default):
    """Returns the :attr:`.Context.clan`"""

    async def default(self, ctx: "commands.Context"):
        return ctx.clan


class DefaultGame(Default):
    """Returns the :attr:`~steam.User.game`"""

    async def default(self, ctx: "commands.Context"):
        return ctx.author.game



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: steammessages_gamenotifications.proto
# plugin: python-betterproto

from dataclasses import dataclass
from typing import List

import betterproto


@dataclass
class CGameNotifications_Variable(betterproto.Message):
    key: str = betterproto.string_field(1)
    value: str = betterproto.string_field(2)


@dataclass
class CGameNotifications_LocalizedText(betterproto.Message):
    token: str = betterproto.string_field(1)
    variables: List["CGameNotifications_Variable"] = betterproto.message_field(2)
    rendered_text: str = betterproto.string_field(3)


@dataclass
class CGameNotifications_UserStatus(betterproto.Message):
    steamid: float = betterproto.fixed64_field(1)
    state: str = betterproto.string_field(2)
    title: "CGameNotifications_LocalizedText" = betterproto.message_field(3)
    message: "CGameNotifications_LocalizedText" = betterproto.message_field(4)


@dataclass
class CGameNotifications_CreateSession_Request(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    context: int = betterproto.uint64_field(2)
    title: "CGameNotifications_LocalizedText" = betterproto.message_field(3)
    users: List["CGameNotifications_UserStatus"] = betterproto.message_field(4)
    steamid: float = betterproto.fixed64_field(5)


@dataclass
class CGameNotifications_CreateSession_Response(betterproto.Message):
    sessionid: int = betterproto.uint64_field(1)


@dataclass
class CGameNotifications_DeleteSession_Request(betterproto.Message):
    sessionid: int = betterproto.uint64_field(1)
    appid: int = betterproto.uint32_field(2)
    steamid: float = betterproto.fixed64_field(3)


@dataclass
class CGameNotifications_DeleteSession_Response(betterproto.Message):
    pass


@dataclass
class CGameNotifications_UpdateSession_Request(betterproto.Message):
    sessionid: int = betterproto.uint64_field(1)
    appid: int = betterproto.uint32_field(2)
    title: "CGameNotifications_LocalizedText" = betterproto.message_field(3)
    users: List["CGameNotifications_UserStatus"] = betterproto.message_field(4)
    steamid: float = betterproto.fixed64_field(6)


@dataclass
class CGameNotifications_UpdateSession_Response(betterproto.Message):
    pass


@dataclass
class CGameNotifications_EnumerateSessions_Request(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    include_all_user_messages: bool = betterproto.bool_field(3)
    include_auth_user_message: bool = betterproto.bool_field(4)
    language: str = betterproto.string_field(5)


@dataclass
class CGameNotifications_Session(betterproto.Message):
    sessionid: int = betterproto.uint64_field(1)
    appid: int = betterproto.uint64_field(2)
    context: int = betterproto.uint64_field(3)
    title: "CGameNotifications_LocalizedText" = betterproto.message_field(4)
    time_created: int = betterproto.uint32_field(5)
    time_updated: int = betterproto.uint32_field(6)
    user_status: List["CGameNotifications_UserStatus"] = betterproto.message_field(7)


@dataclass
class CGameNotifications_EnumerateSessions_Response(betterproto.Message):
    sessions: List["CGameNotifications_Session"] = betterproto.message_field(1)


@dataclass
class CGameNotifications_GetSessionDetails_Request(betterproto.Message):
    sessions: List["CGameNotifications_GetSessionDetails_RequestRequestedSession"] = betterproto.message_field(1)
    appid: int = betterproto.uint32_field(2)
    language: str = betterproto.string_field(3)


@dataclass
class CGameNotifications_GetSessionDetails_RequestRequestedSession(betterproto.Message):
    sessionid: int = betterproto.uint64_field(1)
    include_auth_user_message: bool = betterproto.bool_field(3)


@dataclass
class CGameNotifications_GetSessionDetails_Response(betterproto.Message):
    sessions: List["CGameNotifications_Session"] = betterproto.message_field(1)


@dataclass
class GameNotificationSettings(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    allow_notifications: bool = betterproto.bool_field(2)


@dataclass
class CGameNotifications_UpdateNotificationSettings_Request(betterproto.Message):
    game_notification_settings: List["GameNotificationSettings"] = betterproto.message_field(1)


@dataclass
class CGameNotifications_UpdateNotificationSettings_Response(betterproto.Message):
    pass


@dataclass
class CGameNotifications_OnNotificationsRequested_Notification(betterproto.Message):
    steamid: float = betterproto.fixed64_field(1)
    appid: int = betterproto.uint32_field(2)


@dataclass
class CGameNotifications_OnUserStatusChanged_Notification(betterproto.Message):
    steamid: float = betterproto.fixed64_field(1)
    sessionid: int = betterproto.uint64_field(2)
    appid: int = betterproto.uint32_field(3)
    status: "CGameNotifications_UserStatus" = betterproto.message_field(4)
    removed: bool = betterproto.bool_field(5)



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import asyncio
import json
import logging
import re
from base64 import b64encode
from sys import version_info
from time import time
from typing import TYPE_CHECKING, Any, Awaitable, List, Optional, Tuple, Union

import aiohttp
import rsa
from bs4 import BeautifulSoup
from yarl import URL as _URL

from . import __version__, errors, utils
from .models import URL
from .user import ClientUser

if TYPE_CHECKING:
    from .client import Client
    from .image import Image

log = logging.getLogger(__name__)


async def json_or_text(r: aiohttp.ClientResponse) -> Optional[Any]:
    try:
        return await r.json()
    except aiohttp.ContentTypeError:  # steam is too inconsistent to do this properly
        return await r.text()


class Route:
    def __init__(self, path):
        self.url = _URL(f"{self.BASE}{path}")

    def __str__(self):
        return str(self.url)


class APIRoute(Route):
    BASE = URL.API

    def __init__(self, path):
        path = f'{path}{"/v1" if not path.endswith("v2") else ""}'
        super().__init__(path)


class CRoute(Route):
    BASE = URL.COMMUNITY


class HTTPClient:
    """The HTTP Client that interacts with the Steam web API."""

    SUCCESS_LOG = "{method} {url} has received {text}"
    REQUEST_LOG = "{method} {url} with {payload} has returned {status}"

    def __init__(self, loop: asyncio.AbstractEventLoop, client: "Client"):
        self._loop = loop
        self._session: Optional[aiohttp.ClientSession] = None  # filled in login
        self._client = client

        self.username: Optional[str] = None
        self.password: Optional[str] = None
        self.api_key: Optional[str] = None
        self.shared_secret: Optional[str] = None
        self._one_time_code = ""
        self._captcha_id = "-1"
        self._captcha_text = ""
        self._steam_id = ""

        self.session_id: Optional[str] = None
        self.user: Optional[ClientUser] = None
        self.logged_in = False
        self.user_agent = (
            f"steam.py/{__version__} client (https://github.com/Gobot1234/steam.py), "
            f"Python/{version_info[0]}.{version_info[1]}, aiohttp/{aiohttp.__version__}"
        )

    def recreate(self) -> None:
        if self._session.closed:
            self._session = aiohttp.ClientSession()

    async def request(
        self, method: str, url: Union[APIRoute, CRoute, str], **kwargs
    ) -> Optional[Any]:  # adapted from d.py
        kwargs["headers"] = {"User-Agent": self.user_agent, **kwargs.get("headers", {})}

        for tries in range(5):
            async with self._session.request(method, str(url), **kwargs) as r:
                payload = kwargs.get("data")
                log.debug(
                    self.REQUEST_LOG.format(method=method, url=r.url, payload=f"PAYLOAD: {payload}", status=r.status,)
                )

                # even errors have text involved in them so this is safe to call
                data = await json_or_text(r)

                # the request was successful so just return the text/json
                if 200 <= r.status < 300:
                    log.debug(f"{method} {r.url} has received {data}")
                    return data

                if 300 <= r.status <= 399 and "/login" in r.headers.get("location", ""):  # been logged out
                    log.debug("Logged out of session re-logging in")
                    await self.login(self.username, self.password, self.api_key, self.shared_secret)
                    continue

                # we are being rate limited
                elif r.status == 429:
                    # I haven't been able to get any X-Retry-After headers
                    # from the API but we should probably still handle it
                    try:
                        await asyncio.sleep(float(r.headers["X-Retry-After"]))
                    except KeyError:  # steam being un-helpful as usual
                        await asyncio.sleep(2 ** tries)
                    continue

                # we've received a 500 or 502, an unconditional retry
                elif r.status in {500, 502}:
                    await asyncio.sleep(1 + tries * 3)
                    continue

                elif r.status == 401:
                    # api key either got revoked or it was never valid
                    if not data:
                        raise errors.HTTPException(r, data)
                    if "Access is denied. Retrying will not help. Please verify your <pre>key=</pre>" in data:
                        # time to fetch a new key
                        self._client.api_key = self.api_key = kwargs["key"] = await self.get_api_key()
                        continue
                        # retry with our new key

                # the usual error cases
                elif r.status == 403:
                    raise errors.Forbidden(r, data)
                elif r.status == 404:
                    raise errors.NotFound(r, data)
                else:
                    raise errors.HTTPException(r, data)

        # we've run out of retries, raise
        raise errors.HTTPException(r, data)

    def connect_to_cm(self, cm: str) -> Awaitable:
        headers = {"User-Agent": self.user_agent}
        return self._session.ws_connect(f"wss://{cm}/cmsocket/", timeout=60, headers=headers)

    async def login(self, username: str, password: str, api_key: str, shared_secret: str = None) -> None:
        self.username = username
        self.password = password
        self.shared_secret = shared_secret

        self._session = aiohttp.ClientSession()

        resp = await self._send_login_request()

        if "captcha_needed" in resp:
            self._captcha_id = resp["captcha_gid"]
            self._captcha_text = (
                await utils.ainput(
                    "Please enter the captcha text at"
                    f" https://steamcommunity.com/login/rendercaptcha/?gid={resp['captcha_gid']}"
                )
            ).strip()
            await self.login(username, password, api_key, shared_secret)
        if not resp["success"]:
            raise errors.InvalidCredentials(resp["message"])

        data = resp.get("transfer_parameters")
        if data is None:
            raise errors.LoginError(
                "Cannot perform redirects after login. Steam is likely down, please try again later."
            )

        for url in resp["transfer_urls"]:
            await self.request("POST", url=url, data=data)

        if api_key is None:
            self.api_key = self._client.api_key = await self.get_api_key()
        else:
            self.api_key = self._client.api_key = api_key
        cookies = self._session.cookie_jar.filter_cookies(_URL(URL.COMMUNITY))
        self.session_id = cookies["sessionid"].value

        resp = await self.get_user(resp["transfer_parameters"]["steamid"])
        data = resp["response"]["players"][0]
        self.user = ClientUser(state=self._client._connection, data=data)
        self._client._connection._users[self.user.id64] = self.user
        self.logged_in = True
        self._client.dispatch("login")

    async def logout(self) -> None:
        log.debug("Logging out of session")
        payload = {"sessionid": self.session_id}
        await self.request("POST", CRoute("/login/logout"), data=payload)
        self.logged_in = False
        self._client.dispatch("logout")

    async def _get_rsa_params(self, current_repetitions: int = 0) -> Tuple[bytes, int]:
        payload = {"username": self.username, "donotcache": int(time() * 1000)}
        try:
            key_response = await self.request("POST", CRoute("/login/getrsakey"), data=payload)
        except Exception as e:
            raise errors.LoginError from e
        try:
            rsa_mod = int(key_response["publickey_mod"], 16)
            rsa_exp = int(key_response["publickey_exp"], 16)
            rsa_timestamp = key_response["timestamp"]
        except KeyError:
            if current_repetitions < 5:
                return await self._get_rsa_params(current_repetitions + 1)
            raise ValueError("Could not obtain rsa-key")
        else:
            return b64encode(rsa.encrypt(self.password.encode("utf-8"), rsa.PublicKey(rsa_mod, rsa_exp))), rsa_timestamp

    async def _send_login_request(self) -> dict:
        password, timestamp = await self._get_rsa_params()
        payload = {
            "username": self.username,
            "password": password.decode(),
            "emailauth": self._one_time_code,
            "emailsteamid": self._steam_id,
            "twofactorcode": self._one_time_code,
            "captchagid": self._captcha_id,
            "captcha_text": self._captcha_text,
            "loginfriendlyname": self.user_agent,
            "rsatimestamp": timestamp,
            "remember_login": True,
            "donotcache": int(time() * 100000),
        }
        try:
            resp = await self.request("POST", CRoute("/login/dologin"), data=payload)
            if resp.get("requires_twofactor") or resp.get("emailauth_needed"):
                try:
                    self._steam_id = resp["emailsteamid"]
                except KeyError:
                    pass
                self._one_time_code = await self._client.code()
                return await self._send_login_request()
            return resp
        except Exception as exc:
            raise errors.LoginError from exc

    def get_user(self, user_id64: int) -> Awaitable:
        params = {"key": self.api_key, "steamids": user_id64}
        return self.request("GET", APIRoute("/ISteamUser/GetPlayerSummaries/v2"), params=params)

    async def get_users(self, user_id64s: List[int]) -> List[dict]:
        ret = []

        def chunk():  # chunk the list into 100 element sub-lists for the requests
            for i in range(0, len(user_id64s), 100):
                yield user_id64s[i : i + 100]

        for sublist in chunk():
            for _ in sublist:
                params = {"key": self.api_key, "steamids": ",".join(map(str, sublist))}

            full_resp = await self.request("GET", APIRoute("/ISteamUser/GetPlayerSummaries/v2"), params=params)
            ret.extend(full_resp["response"]["players"])
        return ret

    def add_user(self, user_id64: int) -> Awaitable:
        payload = {
            "sessionID": self.session_id,
            "steamid": user_id64,
            "accept_invite": 0,
        }
        return self.request("POST", CRoute("/actions/AddFriendAjax"), data=payload)

    def remove_user(self, user_id64: int) -> Awaitable:
        payload = {
            "sessionID": self.session_id,
            "steamid": user_id64,
        }
        return self.request("POST", CRoute("/actions/RemoveFriendAjax"), data=payload)

    def block_user(self, user_id64: int) -> Awaitable:
        payload = {"sessionID": self.session_id, "steamid": user_id64, "block": 1}
        return self.request("POST", CRoute("/actions/BlockUserAjax"), data=payload)

    def unblock_user(self, user_id64: int) -> Awaitable:
        payload = {"sessionID": self.session_id, "steamid": user_id64, "block": 0}
        return self.request("POST", CRoute("/actions/BlockUserAjax"), data=payload)

    def accept_user_invite(self, user_id64: int) -> Awaitable:
        payload = {
            "sessionID": self.session_id,
            "steamid": user_id64,
            "accept_invite": 1,
        }
        return self.request("POST", CRoute("/actions/AddFriendAjax"), data=payload)

    def decline_user_invite(self, user_id64: int) -> Awaitable:
        payload = {
            "sessionID": self.session_id,
            "steamid": user_id64,
            "accept_invite": 0,
        }
        return self.request("POST", CRoute("/actions/IgnoreFriendInviteAjax"), data=payload)

    def get_user_games(self, user_id64: int) -> Awaitable:
        params = {
            "key": self.api_key,
            "steamid": user_id64,
            "include_appinfo": 1,
            "include_played_free_games": 1,
        }
        return self.request("GET", APIRoute("/IPlayerService/GetOwnedGames"), params=params)

    def get_user_inventory(self, user_id64: int, app_id: int, context_id: int) -> Awaitable:
        params = {
            "count": 5000,
        }
        return self.request("GET", CRoute(f"/inventory/{user_id64}/{app_id}/{context_id}"), params=params,)

    def get_user_escrow(self, user_id64: int, token: Optional[str]) -> Awaitable:
        params = {
            "key": self.api_key,
            "steamid_target": user_id64,
            "trade_offer_access_token": token if token is not None else "",
        }
        return self.request("GET", APIRoute("/IEconService/GetTradeHoldDurations"), params=params)

    async def get_friends(self, user_id64: int) -> List[dict]:
        params = {"key": self.api_key, "steamid": user_id64, "relationship": "friend"}
        friends = await self.request("GET", APIRoute("/ISteamUser/GetFriendList"), params=params)
        return await self.get_users([friend["steamid"] for friend in friends["friendslist"]["friends"]])

    def get_trade_offers(self, active_only: bool = True, sent: bool = True, received: bool = True) -> Awaitable:
        params = {
            "key": self.api_key,
            "active_only": int(active_only),
            "get_sent_offers": int(sent),
            "get_descriptions": 1,
            "get_received_offers": int(received),
        }
        return self.request("GET", APIRoute("/IEconService/GetTradeOffers"), params=params)

    def get_trade_history(self, limit: int, previous_time: int) -> Awaitable:
        params = {
            "key": self.api_key,
            "max_trades": limit,
            "get_descriptions": 1,
            "include_total": 1,
            "start_after_time": previous_time or 0,
        }
        return self.request("GET", APIRoute("/IEconService/GetTradeHistory"), params=params)

    def get_trade(self, trade_id: int) -> Awaitable:
        params = {"key": self.api_key, "tradeofferid": trade_id, "get_descriptions": 1}
        return self.request("GET", APIRoute("/IEconService/GetTradeOffer"), params=params)

    def accept_user_trade(self, user_id64: int, trade_id: int) -> Awaitable:
        payload = {
            "sessionid": self.session_id,
            "tradeofferid": trade_id,
            "serverid": 1,
            "partner": user_id64,
            "captcha": "",
        }
        headers = {"Referer": f"{URL.COMMUNITY}/tradeoffer/{trade_id}"}
        return self.request("POST", CRoute(f"/tradeoffer/{trade_id}/accept"), data=payload, headers=headers,)

    def decline_user_trade(self, trade_id: int) -> Awaitable:
        payload = {"key": self.api_key, "tradeofferid": trade_id}
        return self.request("POST", APIRoute("/IEconService/DeclineTradeOffer"), data=payload)

    def cancel_user_trade(self, trade_id: int) -> Awaitable:
        payload = {"key": self.api_key, "tradeofferid": trade_id}
        return self.request("POST", APIRoute("/IEconService/CancelTradeOffer"), data=payload)

    def send_trade_offer(
        self,
        user_id64: int,
        user_id: int,
        to_send: List[dict],
        to_receive: List[dict],
        token: Optional[str],
        offer_message: str,
        **kwargs,
    ) -> Awaitable:
        payload = {
            "sessionid": self.session_id,
            "serverid": 1,
            "partner": user_id64,
            "tradeoffermessage": offer_message,
            "json_tradeoffer": json.dumps(
                {
                    "newversion": True,
                    "version": 4,
                    "me": {"assets": to_send, "currency": [], "ready": False},
                    "them": {"assets": to_receive, "currency": [], "ready": False},
                }
            ),
            "captcha": "",
            "trade_offer_create_params": json.dumps({"trade_offer_access_token": token}) if token is not None else {},
        }
        payload.update(**kwargs)
        headers = {"Referer": f"{URL.COMMUNITY}/tradeoffer/new/?partner={user_id}"}
        return self.request("POST", CRoute("/tradeoffer/new/send"), data=payload, headers=headers)

    def send_counter_trade_offer(
        self,
        trade_id: int,
        user_id64: int,
        user_id: int,
        to_send: List[dict],
        to_receive: List[dict],
        token: Optional[str],
        offer_message: str,
    ) -> Awaitable:
        return self.send_trade_offer(user_id64, user_id, to_send, to_receive, token, offer_message, trade_id=trade_id,)

    def get_cm_list(self, cell_id: int) -> Awaitable:
        params = {"cellid": cell_id}
        return self.request("GET", APIRoute("/ISteamDirectory/GetCMList"), params=params)

    def get_comments(self, id64: int, comment_type: str, limit: int = None) -> Awaitable:
        params = {"start": 0, "totalcount": 9999999999}
        if limit is None:
            params["count"] = 9999999999
        else:
            params["count"] = limit
        return self.request("GET", CRoute(f"/comment/{comment_type}/render/{id64}"), params=params)

    def post_comment(self, id64: int, comment_type: str, content: str) -> Awaitable:
        payload = {
            "sessionid": self.session_id,
            "comment": content,
        }
        return self.request("POST", CRoute(f"/comment/{comment_type}/post/{id64}"), data=payload)

    def delete_comment(self, id64: int, comment_id: int, comment_type: str) -> Awaitable:
        payload = {
            "sessionid": self.session_id,
            "gidcomment": comment_id,
        }
        return self.request("POST", CRoute(f"/comment/{comment_type}/delete/{id64}"), data=payload)

    def report_comment(self, id64: int, comment_id: int, comment_type: str) -> Awaitable:
        payload = {"gidcomment": comment_id, "hide": 1}
        return self.request("POST", CRoute(f"/comment/{comment_type}/hideandreport/{id64}"), data=payload,)

    def accept_clan_invite(self, clan_id: int) -> Awaitable:
        payload = {
            "sessionid": self.session_id,
            "steamid": self.user.id64,
            "ajax": "1",
            "action": "group_accept",
            "steamids[]": clan_id,
        }
        return self.request("POST", CRoute("/my/friends/action"), data=payload)

    def decline_clan_invite(self, clan_id: int) -> Awaitable:
        payload = {
            "sessionid": self.session_id,
            "steamid": self.user.id64,
            "ajax": "1",
            "action": "group_ignore",
            "steamids[]": clan_id,
        }
        return self.request("POST", CRoute("/my/friends/action"), data=payload)

    def join_clan(self, clan_id: int) -> Awaitable:
        payload = {
            "sessionID": self.session_id,
            "action": "join",
        }
        return self.request("POST", CRoute(f"/gid/{clan_id}"), data=payload)

    def leave_clan(self, clan_id: int) -> Awaitable:
        payload = {
            "sessionID": self.session_id,
            "action": "leaveGroup",
            "groupId": clan_id,
        }
        return self.request("POST", CRoute("/my/home_process"), data=payload)

    def invite_user_to_clan(self, user_id64: int, clan_id: int) -> Awaitable:
        payload = {
            "sessionID": self.session_id,
            "group": clan_id,
            "invitee": user_id64,
            "type": "groupInvite",
        }
        return self.request("POST", CRoute("/actions/GroupInvite"), data=payload)

    def get_user_clans(self, user_id64: int) -> Awaitable:
        params = {"key": self.api_key, "steamid": user_id64}
        return self.request("GET", APIRoute("/ISteamUser/GetUserGroupList"), params=params)

    def get_user_bans(self, user_id64: int) -> Awaitable:
        params = {"key": self.api_key, "steamids": user_id64}
        return self.request("GET", APIRoute("/ISteamUser/GetPlayerBans"), params=params)

    def get_user_level(self, user_id64: int) -> Awaitable:
        params = {"key": self.api_key, "steamid": user_id64}
        return self.request("GET", APIRoute("/IPlayerService/GetSteamLevel"), params=params)

    def get_user_badges(self, user_id64: int) -> Awaitable:
        params = {"key": self.api_key, "steamid": user_id64}
        return self.request("GET", APIRoute("/IPlayerService/GetBadges"), params=params)

    def clear_nickname_history(self) -> Awaitable:
        payload = {"sessionid": self.session_id}
        return self.request("POST", CRoute("/my/ajaxclearaliashistory"), data=payload)

    def clear_notifications(self) -> Awaitable:
        return self.request("GET", CRoute("/my/inventory"))

    async def edit_profile(
        self, name: str, real_name: str, url: str, summary: str, country: str, state: str, city: str, avatar: "Image",
    ) -> None:
        if any((name, real_name, url, summary, country, state, city, avatar)):
            resp = await self.request("GET", url=CRoute("/my/edit"))
            soup = BeautifulSoup(resp, "html.parser")
            edit_config = str(soup.find("div", attrs={"id": "profile_edit_config"}))
            value = re.findall(r'data-profile-edit=[\'"]{(.*?)},', utils.replace_steam_code(edit_config), flags=re.S,)[
                0
            ]
            loadable = value.replace("\r", "\\r").replace("\n", "\\n")
            profile = json.loads(f'{"{"}{loadable}{"}}"}')
            for key, value in profile.items():
                if isinstance(value, dict):
                    continue
                profile[key] = str(value).replace("\\r", "\r").replace("\\n", "\n")

            payload = {
                "sessionID": self.session_id,
                "type": "profileSave",
                "weblink_1_title": "",
                "weblink_1_url": "",
                "weblink_2_title": "",
                "weblink_2_url": "",
                "weblink_3_title": "",
                "weblink_3_url": "",
                "personaName": name or profile["strPersonaName"],
                "real_name": real_name or profile["strRealName"],
                "customURL": url or profile["strCustomURL"],
                "country": country or profile["LocationData"]["locCountryCode"],
                "state": state or profile["LocationData"]["locStateCode"],
                "city": city or profile["LocationData"]["locCityCode"],
                "summary": summary or profile["strSummary"],
            }

            await self.request("POST", url=f"{self.user.community_url}/edit", data=payload)
        if avatar is not None:
            payload = aiohttp.FormData()
            payload.add_field("MAX_FILE_SIZE", str(len(avatar)))
            payload.add_field("type", "player_avatar_image")
            payload.add_field("sId", str(self.user.id64))
            payload.add_field("sessionid", self.session_id)
            payload.add_field("doSub", "1")
            payload.add_field(
                "avatar", avatar.read(), filename=f"avatar.{avatar.type}", content_type=f"image/{avatar.type}",
            )
            await self.request("POST", CRoute("/actions/FileUploader"), data=payload)

    async def send_user_image(self, user_id64: int, image: "Image") -> None:
        payload = {
            "sessionid": self.session_id,
            "l": "english",
            "file_size": len(image),
            "file_name": image.name,
            "file_sha": image.hash,
            "file_image_width": image.width,
            "file_image_height": image.height,
            "file_type": f"image/{image.type}",
        }
        resp = await self.request("POST", CRoute("/chat/beginfileupload"), data=payload)

        result = resp["result"]
        url = f'{"https" if result["use_https"] else "http"}://{result["url_host"]}{result["url_path"]}'
        headers = {header["name"]: header["value"] for header in result["request_headers"]}
        await self.request("PUT", url=url, headers=headers, data=image.read())

        payload.update(
            {
                "success": 1,
                "ugcid": result["ugcid"],
                "timestamp": resp["timestamp"],
                "hmac": resp["hmac"],
                "friend_steamid": user_id64,
                "spoiler": int(image.spoiler),
            }
        )
        await self.request("POST", CRoute("/chat/commitfileupload"), data=payload)

    async def send_group_image(self, destination: Tuple[int, int], image: "Image") -> None:
        chat_id, channel_id = destination
        payload = {
            "sessionid": self.session_id,
            "l": "english",
            "file_size": len(image),
            "file_name": image.name,
            "file_sha": image.hash,
            "file_image_width": image.width,
            "file_image_height": image.height,
            "file_type": f"image/{image.type}",
        }
        resp = await self.request("POST", CRoute("/chat/beginfileupload"), data=payload)

        result = resp["result"]
        url = f'{"https" if result["use_https"] else "http"}://{result["url_host"]}{result["url_path"]}'
        headers = {header["name"]: header["value"] for header in result["request_headers"]}
        await self.request("PUT", url=url, headers=headers, data=image.read())

        payload.update(
            {
                "success": 1,
                "ugcid": result["ugcid"],
                "timestamp": resp["timestamp"],
                "hmac": resp["hmac"],
                "chat_group_id": channel_id,
                "chat_id": chat_id,
                "spoiler": int(image.spoiler),
            }
        )
        await self.request("POST", CRoute("/chat/commitfileupload"), data=payload)

    async def get_api_key(self) -> str:
        resp = await self.request("GET", CRoute("/dev/apikey"))
        if "<h2>Access Denied</h2>" in resp:
            raise errors.LoginError(
                "Access denied, you will need to generate a key yourself: https://steamcommunity.com/dev/apikey"
            )
        error = "You must have a validated email address to create a Steam Web API key"
        if error in resp:
            raise errors.LoginError(error)

        match = re.findall(r"<p>Key: ([0-9A-F]+)</p>", resp)
        if match:
            return match[0]

        self.session_id = re.findall(r'g_sessionID = "(.*?)";', resp)[0]
        payload = {
            "domain": "steam.py",
            "agreeToTerms": "agreed",
            "sessionid": self.session_id,
            "Submit": "Register",
        }
        resp = await self.request("POST", CRoute("/dev/registerkey"), data=payload)
        return re.findall(r"<p>Key: ([0-9A-F]+)</p>", resp)[0]



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2015 Rossen Georgiev <rossen@rgp.io>
Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

This contains a copy of
https://github.com/ValvePython/steam/blob/master/steam/steamid.py
"""

import abc
import asyncio
import re
from datetime import datetime
from typing import TYPE_CHECKING, Awaitable, Callable, List, Optional, Tuple, Union

from .badge import UserBadges
from .comment import Comment
from .enums import EInstanceFlag, EPersonaState, EPersonaStateFlag, EResult, EType, ETypeChar, EUniverse
from .errors import WSException
from .game import Game
from .iterators import CommentsIterator
from .models import URL, Ban
from .trade import Inventory
from .utils import _INVITE_HEX, _INVITE_MAPPING, make_steam64, steam64_from_url

if TYPE_CHECKING:
    from aiohttp import ClientSession

    from .clan import Clan
    from .client import EventType
    from .group import Group
    from .image import Image
    from .state import ConnectionState
    from .user import User


__all__ = (
    "SteamID",
    "Message",
)


class SteamID(metaclass=abc.ABCMeta):
    """Convert a Steam ID between its various representations."""

    __slots__ = ("_BASE", "__weakref__")

    def __init__(self, *args, **kwargs):
        self._BASE = make_steam64(*args, **kwargs)

    def __repr__(self):
        attrs = ("id", "type", "universe", "instance")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<SteamID {' '.join(resolved)}>"

    def __int__(self):
        return self._BASE

    def __bool__(self):
        return bool(self._BASE)

    def __str__(self):
        return str(self._BASE)

    def __hash__(self):
        return hash(self._BASE)

    def __eq__(self, other):
        return self._BASE == other

    @property
    def id(self) -> int:
        """:class:`int`: Represents the account id.
        This is also known as the 32 bit id.
        """
        return int(self) & 0xFFFFFFFF

    @property
    def instance(self) -> int:
        """:class:`int`: Returns the instance of the account."""
        return (int(self) >> 32) & 0xFFFFF

    @property
    def type(self) -> EType:
        """:class:`~steam.EType`: Represents the Steam type of the account."""
        return EType((int(self) >> 52) & 0xF)

    @property
    def universe(self) -> EUniverse:
        """:class:`~steam.EUniverse`: Represents the Steam universe of the account."""
        return EUniverse((int(self) >> 56) & 0xFF)

    @property
    def as_32(self) -> int:
        """:class:`int`: The account's id.
        An alias to :attr:`SteamID.id`.
        """
        return self.id

    @property
    def id64(self) -> int:
        """:class:`int`: The steam 64 bit id of the account.
        Used for community profiles along with other useful things.
        """
        return int(self)

    @property
    def as_64(self) -> int:
        """:class:`int`: The steam 64 bit id of the account.
        An alias to :attr:`id64`.
        """
        return self.id64

    @property
    def id2(self) -> str:
        """:class:`str`: The Steam2 id of the account.
            e.g ``STEAM_1:0:1234``.

        .. note::
            ``STEAM_X:Y:Z``. The value of ``X`` should represent the universe, or ``1``
            for ``Public``. However, there was a bug in GoldSrc and Orange Box games
            and ``X`` was ``0``. If you need that format use :attr:`as_steam2_zero`.
        """
        return f"STEAM_{int(self.universe)}:{self.id % 2}:{self.id >> 1}"

    @property
    def as_steam2(self) -> str:
        """:class:`str`: The Steam2 id of the account.
            e.g ``STEAM_1:0:1234``.

        .. note::
            ``STEAM_X:Y:Z``. The value of ``X`` should represent the universe, or ``1``
            for ``Public``. However, there was a bug in GoldSrc and Orange Box games
            and ``X`` was ``0``. If you need that format use :attr:`as_steam2_zero`.

        An alias to :attr:`id2`.
        """
        return self.id2

    @property
    def as_steam2_zero(self) -> str:
        """:class:`str`: The Steam2 id of the account.
            e.g ``STEAM_0:0:1234``.

        For GoldSrc and Orange Box games.
        See :attr:`id2`.
        """
        return self.as_steam2.replace("_1", "_0")

    @property
    def id3(self) -> str:
        """:class:`str`: The Steam3 id of the account.
            e.g ``[U:1:1234]``.

        This is used for more recent games.
        """
        typechar = ETypeChar(self.type).name
        instance = None

        if self.type in (EType.AnonGameServer, EType.Multiseat):
            instance = self.instance
        elif self.type == EType.Individual:
            if self.instance != 1:
                instance = self.instance
        elif self.type == EType.Chat:
            if self.instance & EInstanceFlag.Clan:
                typechar = "c"
            elif self.instance & EInstanceFlag.Lobby:
                typechar = "L"
            else:
                typechar = "T"

        parts = [typechar, int(self.universe), self.id]

        if instance is not None:
            parts.append(instance)

        return f'[{":".join(map(str, parts))}]'

    @property
    def as_steam3(self) -> str:
        """:class:`str`: The Steam3 id of the account.
        An alias to :attr:`id3`.
        """
        return self.id3

    @property
    def invite_code(self) -> Optional[str]:
        """Optional[:class:`str`]: s.team invite code format.
        e.g. ``cv-dgb``
        """
        if self.type == EType.Individual and self.is_valid():

            def repl_mapper(x):
                return _INVITE_MAPPING[x.group()]

            invite_code = re.sub(f"[{_INVITE_HEX}]", repl_mapper, f"{self.id:x}")
            split_idx = len(invite_code) // 2

            if split_idx:
                invite_code = f"{invite_code[:split_idx]}-{invite_code[split_idx:]}"

            return invite_code

    @property
    def invite_url(self) -> Optional[str]:
        """Optional[:class:`str`]: The user's full invite code URL.
        e.g ``https://s.team/p/cv-dgb``
        """
        code = self.invite_code
        if code:
            return f"https://s.team/p/{code}"

    @property
    def community_url(self) -> Optional[str]:
        """Optional[:class:`str`]: The community url of the account
        e.g https://steamcommunity.com/profiles/123456789.
        """
        suffix = {
            EType.Individual: "profiles",
            EType.Clan: "gid",
        }
        if self.type in suffix:
            return f"https://steamcommunity.com/{suffix[self.type]}/{self.id64}"

    def is_valid(self) -> bool:
        """:class:`bool`: Check whether this SteamID is valid.
        This doesn't however mean that a matching profile can be found.
        """
        if self.type == EType.Invalid or self.type >= EType.Max:
            return False

        if self.universe == EUniverse.Invalid or self.universe >= EUniverse.Max:
            return False

        if self.type == EType.Individual:
            if self.id == 0 or self.instance > 4:
                return False

        if self.type == EType.Clan:
            if self.id == 0 or self.instance != 0:
                return False

        if self.type == EType.GameServer:
            if self.id == 0:
                return False

        if self.type == EType.AnonGameServer:
            if self.id == 0 and self.instance == 0:
                return False

        return True

    @classmethod
    async def from_url(cls, url: str, session: "ClientSession" = None, timeout: float = 30) -> Optional["SteamID"]:
        """Takes Steam community url and returns a SteamID instance or ``None``.
        See :func:`steam64_from_url` for details.

        Parameters
        ----------
        url: :class:`str`
            The Steam community url.
        session: Optional[:class:`aiohttp.ClientSession`]
            The session to make the request with. If
            ``None`` is passed a new one is generated.
        timeout: Optional[:class:`float`]
            How long to wait on http request before turning ``None``.

        Returns
        -------
        Optional[:class:`SteamID`]
            `SteamID` instance or ``None``.
        """
        id64 = await steam64_from_url(url, session, timeout)
        return cls(id64) if id64 else None


class BaseUser(SteamID):
    """An ABC that details the common operations on a Steam user.
    The following classes implement this ABC:

        - :class:`~steam.User`
        - :class:`~steam.ClientUser`

    .. container:: operations

        .. describe:: x == y

            Checks if two users are equal.

        .. describe:: x != y

            Checks if two users are not equal.

        .. describe:: str(x)

            Returns the user's name.

    Attributes
    ----------
    name: :class:`str`
        The user's username.
    state: :class:`~steam.EPersonaState`
        The current persona state of the account (e.g. LookingToTrade).
    game: Optional[:class:`~steam.Game`]
        The Game instance attached to the user. Is None if the user
        isn't in a game or one that is recognised by the api.
    primary_clan: Optional[:class:`SteamID`]
        The primary clan the User displays on their profile.
    avatar_url: :class:`str`
        The avatar url of the user. Uses the large (184x184 px) image url.
    real_name: Optional[:class:`str`]
        The user's real name defined by them. Could be ``None``.
    created_at: Optional[:class:`datetime.datetime`]
        The time at which the user's account was created. Could be ``None``.
    last_logoff: Optional[:class:`datetime.datetime`]
        The last time the user logged into steam. Could be None (e.g. if they are currently online).
    country: Optional[:class:`str`]
        The country code of the account. Could be ``None``.
    flags: Union[:class:`~steam.EPersonaStateFlag`, :class:`int`]
        The persona state flags of the account.
    """

    __slots__ = (
        "name",
        "game",
        "state",
        "flags",
        "country",
        "primary_clan",
        "trade_url",
        "real_name",
        "avatar_url",
        "last_seen_online",
        "created_at",
        "last_logoff",
        "last_logon",
        "_state",
        "_data",
    )

    def __init__(self, state: "ConnectionState", data: dict):
        super().__init__(data["steamid"])
        self._state = state
        self.name: Optional[str] = None
        self.real_name: Optional[str] = None
        self.avatar_url: Optional[str] = None
        self.primary_clan: Optional[SteamID] = None
        self.country: Optional[str] = None
        self.created_at: Optional[datetime] = None
        self.last_logoff: Optional[datetime] = None
        self.last_logon: Optional[datetime] = None
        self.last_seen_online: Optional[datetime] = None
        self.state: Optional[EPersonaState] = None
        self.flags: Optional[EPersonaStateFlag] = None
        self.game: Optional[Game] = None
        self._update(data)

    def __repr__(self):
        attrs = ("name", "state", "id", "type", "universe", "instance")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<User {' '.join(resolved)}>"

    def __str__(self):
        return self.name

    def _update(self, data) -> None:
        self._data = data
        self.name = data["personaname"]
        self.real_name = data.get("realname") or self.real_name
        self.avatar_url = data.get("avatarfull") or self.avatar_url
        self.trade_url = f"{URL.COMMUNITY}/tradeoffer/new/?partner={self.id}"

        self.primary_clan = SteamID(data["primaryclanid"]) if "primaryclanid" in data else None or self.primary_clan
        self.country = data.get("loccountrycode") or self.country
        self.created_at = (
            datetime.utcfromtimestamp(data["timecreated"]) if "timecreated" in data else None or self.created_at
        )
        self.last_logoff = (
            datetime.utcfromtimestamp(data["lastlogoff"]) if "lastlogoff" in data else None or self.last_logoff
        )
        self.last_logon = (
            datetime.utcfromtimestamp(data["last_logon"]) if "last_logon" in data else None or self.last_logon
        )
        self.last_seen_online = (
            datetime.utcfromtimestamp(data["last_seen_online"])
            if "last_seen_online" in data
            else None or self.last_seen_online
        )
        self.game = (
            Game(title=data.get("gameextrainfo"), app_id=data["gameid"]) if "gameid" in data else None or self.game
        )
        self.state = EPersonaState(data.get("personastate", 0)) or self.state
        self.flags = EPersonaStateFlag.try_value(data.get("personastateflags", 0)) or self.flags

    async def comment(self, content: str) -> Comment:
        """|coro|
        Post a comment to an :class:`User`'s profile.

        Parameters
        -----------
        content: :class:`str`
            The message to add to the user's profile.

        Returns
        -------
        :class:`~steam.Comment`
            The created comment.
        """
        resp = await self._state.http.post_comment(self.id64, "Profile", content)
        id = int(re.findall(r'id="comment_(\d+)"', resp["comments_html"])[0])
        timestamp = datetime.utcfromtimestamp(resp["timelastpost"])
        comment = Comment(
            state=self._state, id=id, owner=self, timestamp=timestamp, content=content, author=self._state.client.user,
        )
        self._state.dispatch("comment", comment)
        return comment

    async def inventory(self, game: Game) -> Inventory:
        """|coro|
        Fetch an :class:`User`'s :class:`~steam.Inventory` for trading.

        Parameters
        -----------
        game: :class:`~steam.Game`
            The game to fetch the inventory for.

        Raises
        ------
        :exc:`~steam.Forbidden`
            The user's inventory is private.

        Returns
        -------
        :class:`Inventory`
            The user's inventory.
        """
        resp = await self._state.http.get_user_inventory(self.id64, game.app_id, game.context_id)
        return Inventory(state=self._state, data=resp, owner=self)

    async def friends(self) -> List["User"]:
        """|coro|
        Fetch the list of :class:`~steam.User`'s friends from the API.

        Returns
        -------
        List[:class:`~steam.User`]
            The list of user's friends from the API.
        """
        friends = await self._state.http.get_friends(self.id64)
        return [self._state._store_user(friend) for friend in friends]

    async def games(self) -> List[Game]:
        """|coro|
        Fetches the list of :class:`~steam.Game`
        objects the :class:`User` owns from the API.

        Returns
        -------
        List[:class:`~steam.Game`]
            The list of game objects from the API.
        """
        data = await self._state.http.get_user_games(self.id64)
        games = data["response"].get("games", [])
        return [Game._from_api(game) for game in games]

    async def clans(self) -> List["Clan"]:
        """|coro|
        Fetches a list of the :class:`User`'s :class:`~steam.Clan`
        objects the :class:`User` is in from the API.

        Returns
        -------
        List[:class:`~steam.Clan`]
            The user's clans.
        """
        clans = []

        async def getter(gid: int):
            try:
                clan = await self._state.client.fetch_clan(gid)
            except WSException as exc:
                if exc.code == (EResult.LimitExceeded, EResult.RateLimitExceeded):
                    await asyncio.sleep(20)
                    await getter(gid)
            else:
                clans.append(clan)

        resp = await self._state.http.get_user_clans(self.id64)
        for clan in resp["response"]["groups"]:
            await getter(int(clan["gid"]))
        return clans

    async def bans(self) -> Ban:
        """|coro|
        Fetches the :class:`User`'s :class:`~steam.Ban` objects.

        Returns
        -------
        :class:`~steam.Ban`
            The user's bans.
        """
        resp = await self._state.http.get_user_bans(self.id64)
        resp = resp["players"][0]
        resp["EconomyBan"] = False if resp["EconomyBan"] == "none" else True
        return Ban(data=resp)

    async def badges(self) -> UserBadges:
        """|coro|
        Fetches the :class:`User`'s :class:`~steam.UserBadges` objects.

        Returns
        -------
        :class:`~steam.UserBadges`
            The user's badges.
        """
        resp = await self._state.http.get_user_badges(self.id64)
        return UserBadges(data=resp["response"])

    async def level(self) -> int:
        """|coro|
        Fetches the :class:`User`'s level.

        Returns
        -------
        :class:`int`
            The user's level.
        """
        resp = await self._state.http.get_user_level(self.id64)
        return resp["response"]["player_level"]

    def is_commentable(self) -> bool:
        """:class:`bool`: Specifies if the user's account is able to be commented on."""
        return bool(self._data.get("commentpermission"))

    def is_private(self) -> bool:
        """:class:`bool`: Specifies if the user has a public profile."""
        state = self._data.get("communityvisibilitystate", 0)
        return state in (0, 1, 2)

    def has_setup_profile(self) -> bool:
        """:class:`bool`: Specifies if the user has a setup their profile."""
        return bool(self._data.get("profilestate"))

    async def is_banned(self) -> bool:
        """|coro|
        Specifies if the user is banned from any part of Steam.

        This is equivalent to: ::

            bans = await user.bans()
            bans.is_banned()

        Returns
        -------
        :class:`bool`
            Whether or not the user is banned.
        """
        bans = await self.bans()
        return bans.is_banned()

    def comments(
        self, limit: Optional[int] = None, before: datetime = None, after: datetime = None,
    ) -> CommentsIterator:
        """An :class:`~steam.iterators.AsyncIterator` for accessing a
        :class:`~steam.User`'s :class:`~steam.Comment` objects.

        Examples
        -----------

        Usage: ::

            async for comment in user.comments(limit=10):
                print('Author:', comment.author, 'Said:', comment.content)

        Flattening into a list: ::

            comments = await user.comments(limit=50).flatten()
            # comments is now a list of Comment

        All parameters are optional.

        Parameters
        ----------
        limit: Optional[:class:`int`]
            The maximum number of comments to search through.
            Default is ``None`` which will fetch the user's entire comments section.
        before: Optional[:class:`datetime.datetime`]
            A time to search for comments before.
        after: Optional[:class:`datetime.datetime`]
            A time to search for comments after.

        Yields
        ---------
        :class:`~steam.Comment`
            The comment with the comment information parsed.
        """
        return CommentsIterator(state=self._state, owner=self, limit=limit, before=before, after=after)


_EndPointReturnType = Tuple[Union[Tuple[int, int], int], "EventType"]


class Messageable(metaclass=abc.ABCMeta):
    """An ABC that details the common operations on a Steam message.
    The following classes implement this ABC:

        - :class:`~steam.User`
        - :class:`~steam.ClanChannel`
        - :class:`~steam.GroupChannel`
        - :class:`~steam.DMChannel`
    """

    __slots__ = ()

    def _get_message_endpoint(self) -> _EndPointReturnType:
        pass

    def _get_image_endpoint(self) -> _EndPointReturnType:
        pass

    async def send(self, content: str = None, image: "Image" = None):
        """|coro|
        Send a message to a certain destination.

        Parameters
        ----------
        content: Optional[:class:`str`]
            The content of the message to send.
        image: Optional[:class:`.Image`]
            The image to send to the user.

        Raises
        ------
        :exc:`~steam.HTTPException`
            Sending the message failed.
        :exc:`~steam.Forbidden`
            You do not have permission to send the message.
        """
        if content is not None:
            destination, message_func = self._get_message_endpoint()
            await message_func(destination, str(content))
        if image is not None:
            destination, image_func = self._get_image_endpoint()
            await image_func(destination, image)


class BaseChannel(Messageable):
    __slots__ = ("_state", "participant", "clan", "group")

    def __init__(self):
        self._state: "ConnectionState"
        self.participant: Optional["BaseUser"] = None
        self.clan: Optional["Clan"] = None
        self.group: Optional["Group"] = None

    def typing(self):
        pass

    async def trigger_typing(self):
        pass


class Message:
    """Represents a message from a :class:`~steam.User`
    This is a base class from which all messages inherit.

    The following classes implement this ABC:

        - :class:`~steam.UserMessage`
        - :class:`~steam.GroupMessage`
        - :class:`~steam.ClanMessage`

    Attributes
    ----------
    channel: :class:`steam.abc.BaseChannel`
        The channel the message was sent in.
    content: :class:`str`
        The message's content.
    author: :class:`steam.abc.BaseUser`
        The message's author.
    created_at: :class:`datetime.datetime`
        The time the message was sent at.
    """

    __slots__ = (
        "author",
        "content",
        "channel",
        "created_at",
        "group",
        "clan",
        "_state",
    )

    def __init__(self, channel: "BaseChannel"):
        self._state: "ConnectionState" = channel._state
        self.channel = channel
        self.content: Optional[str] = None
        self.author: Optional[BaseUser] = None
        self.created_at: Optional[datetime] = None
        self.group = channel.group
        self.clan = channel.clan



# -*- coding: utf-8 -*-

from . import (
    steammessages_chat,
    steammessages_econ,
    steammessages_friendmessages,
    steammessages_gameservers,
    steammessages_player,
    steammessages_twofactor,
    steammessages_useraccount,
)

__all__ = ("UMS",)


UMS = {
    "GameServers.GetServerList#1_Request": steammessages_gameservers.CGameServers_GetServerList_Request,
    "GameServers.GetServerList#1_Response": steammessages_gameservers.CGameServers_GetServerList_Response,
    "GameServers.GetServerSteamIDsByIP#1_Request": steammessages_gameservers.CGameServers_GetServerSteamIDsByIP_Request,
    "GameServers.GetServerSteamIDsByIP#1_Response": steammessages_gameservers.CGameServers_IPsWithSteamIDs_Response,
    "GameServers.GetServerIPsBySteamID#1_Request": steammessages_gameservers.CGameServers_GetServerIPsBySteamID_Request,
    "GameServers.GetServerIPsBySteamID#1_Response": steammessages_gameservers.CGameServers_IPsWithSteamIDs_Response,
    "TwoFactor.AddAuthenticator#1_Request": steammessages_twofactor.CTwoFactor_AddAuthenticator_Request,
    "TwoFactor.AddAuthenticator#1_Response": steammessages_twofactor.CTwoFactor_AddAuthenticator_Response,
    "TwoFactor.FinalizeAddAuthenticator#1_Request": steammessages_twofactor.CTwoFactor_FinalizeAddAuthenticator_Request,
    "TwoFactor.FinalizeAddAuthenticator#1_Response": steammessages_twofactor.CTwoFactor_FinalizeAddAuthenticator_Response,
    "TwoFactor.SendEmail#1_Request": steammessages_twofactor.CTwoFactor_SendEmail_Request,
    "TwoFactor.SendEmail#1_Response": steammessages_twofactor.CTwoFactor_SendEmail_Response,
    "TwoFactor.RemoveAuthenticator#1_Request": steammessages_twofactor.CTwoFactor_RemoveAuthenticator_Request,
    "TwoFactor.RemoveAuthenticator#1_Response": steammessages_twofactor.CTwoFactor_RemoveAuthenticator_Response,
    "Player.GetGameBadgeLevels#1_Request": steammessages_player.CPlayer_GetGameBadgeLevels_Request,
    "Player.GetGameBadgeLevels#1_Response": steammessages_player.CPlayer_GetGameBadgeLevels_Response,
    "Player.GetNicknameList#1_Request": steammessages_player.CPlayer_GetNicknameList_Request,
    "Player.GetNicknameList#1_Response": steammessages_player.CPlayer_GetNicknameList_Response,
    "Player.GetEmoticonList#1_Request": steammessages_player.CPlayer_GetEmoticonList_Request,
    "Player.GetEmoticonList#1_Response": steammessages_player.CPlayer_GetEmoticonList_Response,
    "Player.GetPrivacySettings#1_Request": steammessages_player.CPlayer_GetPrivacySettings_Request,
    "Player.GetPrivacySettings#1_Response": steammessages_player.CPlayer_GetPrivacySettings_Response,
    "Player.GetOwnedGames#1_Request": steammessages_player.CPlayer_GetLastPlayedTimes_Request,
    "Player.GetOwnedGames#1_Response": steammessages_player.CPlayer_GetLastPlayedTimes_Response,
    "Econ.GetAssetClassInfo#1_Request": steammessages_econ.CEcon_GetAssetClassInfo_Request,
    "Econ.GetAssetClassInfo#1_Response": steammessages_econ.CEcon_GetAssetClassInfo_Response,
    "Econ.GetTradeOfferAccessToken#1_Request": steammessages_econ.CEcon_GetTradeOfferAccessToken_Request,
    "Econ.GetTradeOfferAccessToken#1_Response": steammessages_econ.CEcon_GetTradeOfferAccessToken_Response,
    "ChatRoom.CreateChatRoomGroup#1_Request": steammessages_chat.CChatRoom_CreateChatRoomGroup_Request,
    "ChatRoom.CreateChatRoomGroup#1_Response": steammessages_chat.CChatRoom_CreateChatRoomGroup_Response,
    "ChatRoom.SaveChatRoomGroup#1_Request": steammessages_chat.CChatRoom_SaveChatRoomGroup_Request,
    "ChatRoom.SaveChatRoomGroup#1_Response": steammessages_chat.CChatRoom_SaveChatRoomGroup_Response,
    "ChatRoom.RenameChatRoomGroup#1_Request": steammessages_chat.CChatRoom_RenameChatRoomGroup_Request,
    "ChatRoom.RenameChatRoomGroup#1_Response": steammessages_chat.CChatRoom_RenameChatRoomGroup_Response,
    "ChatRoom.SetChatRoomGroupTagline#1_Request": steammessages_chat.CChatRoom_SetChatRoomGroupTagline_Request,
    "ChatRoom.SetChatRoomGroupTagline#1_Response": steammessages_chat.CChatRoom_SetChatRoomGroupTagline_Response,
    "ChatRoom.SetChatRoomGroupAvatar#1_Request": steammessages_chat.CChatRoom_SetChatRoomGroupAvatar_Request,
    "ChatRoom.SetChatRoomGroupAvatar#1_Response": steammessages_chat.CChatRoom_SetChatRoomGroupAvatar_Response,
    "ChatRoom.MuteUserInGroup#1_Request": steammessages_chat.CChatRoom_MuteUser_Request,
    "ChatRoom.MuteUserInGroup#1_Response": steammessages_chat.CChatRoom_MuteUser_Response,
    "ChatRoom.KickUserFromGroup#1_Request": steammessages_chat.CChatRoom_KickUser_Request,
    "ChatRoom.KickUserFromGroup#1_Response": steammessages_chat.CChatRoom_KickUser_Response,
    "ChatRoom.SetUserBanState#1_Request": steammessages_chat.CChatRoom_SetUserBanState_Request,
    "ChatRoom.SetUserBanState#1_Response": steammessages_chat.CChatRoom_SetUserBanState_Response,
    "ChatRoom.RevokeInviteToGroup#1_Request": steammessages_chat.CChatRoom_RevokeInvite_Request,
    "ChatRoom.RevokeInviteToGroup#1_Response": steammessages_chat.CChatRoom_RevokeInvite_Response,
    "ChatRoom.CreateRole#1_Request": steammessages_chat.CChatRoom_CreateRole_Request,
    "ChatRoom.CreateRole#1_Response": steammessages_chat.CChatRoom_CreateRole_Response,
    "ChatRoom.GetRoles#1_Request": steammessages_chat.CChatRoom_GetRoles_Request,
    "ChatRoom.GetRoles#1_Response": steammessages_chat.CChatRoom_GetRoles_Response,
    "ChatRoom.RenameRole#1_Request": steammessages_chat.CChatRoom_RenameRole_Request,
    "ChatRoom.RenameRole#1_Response": steammessages_chat.CChatRoom_RenameRole_Response,
    "ChatRoom.ReorderRole#1_Request": steammessages_chat.CChatRoom_ReorderRole_Request,
    "ChatRoom.ReorderRole#1_Response": steammessages_chat.CChatRoom_ReorderRole_Response,
    "ChatRoom.DeleteRole#1_Request": steammessages_chat.CChatRoom_DeleteRole_Request,
    "ChatRoom.DeleteRole#1_Response": steammessages_chat.CChatRoom_DeleteRole_Response,
    "ChatRoom.GetRoleActions#1_Request": steammessages_chat.CChatRoom_GetRoleActions_Request,
    "ChatRoom.GetRoleActions#1_Response": steammessages_chat.CChatRoom_GetRoleActions_Response,
    "ChatRoom.ReplaceRoleActions#1_Request": steammessages_chat.CChatRoom_ReplaceRoleActions_Request,
    "ChatRoom.ReplaceRoleActions#1_Response": steammessages_chat.CChatRoom_ReplaceRoleActions_Response,
    "ChatRoom.AddRoleToUser#1_Request": steammessages_chat.CChatRoom_AddRoleToUser_Request,
    "ChatRoom.AddRoleToUser#1_Response": steammessages_chat.CChatRoom_AddRoleToUser_Response,
    "ChatRoom.GetRolesForUser#1_Request": steammessages_chat.CChatRoom_GetRolesForUser_Request,
    "ChatRoom.GetRolesForUser#1_Response": steammessages_chat.CChatRoom_GetRolesForUser_Response,
    "ChatRoom.DeleteRoleFromUser#1_Request": steammessages_chat.CChatRoom_DeleteRoleFromUser_Request,
    "ChatRoom.DeleteRoleFromUser#1_Response": steammessages_chat.CChatRoom_DeleteRoleFromUser_Response,
    "ChatRoom.JoinChatRoomGroup#1_Request": steammessages_chat.CChatRoom_JoinChatRoomGroup_Request,
    "ChatRoom.JoinChatRoomGroup#1_Response": steammessages_chat.CChatRoom_JoinChatRoomGroup_Response,
    "ChatRoom.InviteFriendToChatRoomGroup#1_Request": steammessages_chat.CChatRoom_InviteFriendToChatRoomGroup_Request,
    "ChatRoom.InviteFriendToChatRoomGroup#1_Response": steammessages_chat.CChatRoom_InviteFriendToChatRoomGroup_Response,
    "ChatRoom.LeaveChatRoomGroup#1_Request": steammessages_chat.CChatRoom_LeaveChatRoomGroup_Request,
    "ChatRoom.LeaveChatRoomGroup#1_Response": steammessages_chat.CChatRoom_LeaveChatRoomGroup_Response,
    "ChatRoom.CreateChatRoom#1_Request": steammessages_chat.CChatRoom_CreateChatRoom_Request,
    "ChatRoom.CreateChatRoom#1_Response": steammessages_chat.CChatRoom_CreateChatRoom_Response,
    "ChatRoom.DeleteChatRoom#1_Request": steammessages_chat.CChatRoom_DeleteChatRoom_Request,
    "ChatRoom.DeleteChatRoom#1_Response": steammessages_chat.CChatRoom_DeleteChatRoom_Response,
    "ChatRoom.RenameChatRoom#1_Request": steammessages_chat.CChatRoom_RenameChatRoom_Request,
    "ChatRoom.RenameChatRoom#1_Response": steammessages_chat.CChatRoom_RenameChatRoom_Response,
    "ChatRoom.SendChatMessage#1_Request": steammessages_chat.CChatRoom_SendChatMessage_Request,
    "ChatRoom.SendChatMessage#1_Response": steammessages_chat.CChatRoom_SendChatMessage_Response,
    "ChatRoom.JoinVoiceChat#1_Request": steammessages_chat.CChatRoom_JoinVoiceChat_Request,
    "ChatRoom.JoinVoiceChat#1_Response": steammessages_chat.CChatRoom_JoinVoiceChat_Response,
    "ChatRoom.LeaveVoiceChat#1_Request": steammessages_chat.CChatRoom_LeaveVoiceChat_Request,
    "ChatRoom.LeaveVoiceChat#1_Response": steammessages_chat.CChatRoom_LeaveVoiceChat_Response,
    "ChatRoom.GetMessageHistory#1_Request": steammessages_chat.CChatRoom_GetMessageHistory_Request,
    "ChatRoom.GetMessageHistory#1_Response": steammessages_chat.CChatRoom_GetMessageHistory_Response,
    "ChatRoom.GetMyChatRoomGroups#1_Request": steammessages_chat.CChatRoom_GetMyChatRoomGroups_Request,
    "ChatRoom.GetMyChatRoomGroups#1_Response": steammessages_chat.CChatRoom_GetMyChatRoomGroups_Response,
    "ChatRoom.GetChatRoomGroupState#1_Request": steammessages_chat.CChatRoom_GetChatRoomGroupState_Request,
    "ChatRoom.GetChatRoomGroupState#1_Response": steammessages_chat.CChatRoom_GetChatRoomGroupState_Response,
    "ChatRoom.GetChatRoomGroupSummary#1_Request": steammessages_chat.CChatRoom_GetChatRoomGroupSummary_Request,
    "ChatRoom.GetChatRoomGroupSummary#1_Response": steammessages_chat.CChatRoom_GetChatRoomGroupSummary_Response,
    "ChatRoom.AckChatMessage#1": steammessages_chat.CChatRoom_AckChatMessage_Notification,
    "ChatRoom.CreateInviteLink#1_Request": steammessages_chat.CChatRoom_CreateInviteLink_Request,
    "ChatRoom.CreateInviteLink#1_Response": steammessages_chat.CChatRoom_CreateInviteLink_Response,
    "ChatRoom.GetInviteLinkInfo#1_Request": steammessages_chat.CChatRoom_GetInviteLinkInfo_Request,
    "ChatRoom.GetInviteLinkInfo#1_Response": steammessages_chat.CChatRoom_GetInviteLinkInfo_Response,
    "ChatRoom.GetInviteInfo#1_Request": steammessages_chat.CChatRoom_GetInviteInfo_Request,
    "ChatRoom.GetInviteInfo#1_Response": steammessages_chat.CChatRoom_GetInviteInfo_Response,
    "ChatRoom.GetInviteLinksForGroup#1_Request": steammessages_chat.CChatRoom_GetInviteLinksForGroup_Request,
    "ChatRoom.GetInviteLinksForGroup#1_Response": steammessages_chat.CChatRoom_GetInviteLinksForGroup_Response,
    "ChatRoom.GetBanList#1_Request": steammessages_chat.CChatRoom_GetBanList_Request,
    "ChatRoom.GetBanList#1_Response": steammessages_chat.CChatRoom_GetBanList_Response,
    "ChatRoom.GetInviteList#1_Request": steammessages_chat.CChatRoom_GetInviteList_Request,
    "ChatRoom.GetInviteList#1_Response": steammessages_chat.CChatRoom_GetInviteList_Response,
    "ChatRoom.DeleteInviteLink#1_Request": steammessages_chat.CChatRoom_DeleteInviteLink_Request,
    "ChatRoom.DeleteInviteLink#1_Response": steammessages_chat.CChatRoom_DeleteInviteLink_Response,
    "ChatRoom.SetSessionActiveChatRoomGroups#1_Request": steammessages_chat.CChatRoom_SetSessionActiveChatRoomGroups_Request,
    "ChatRoom.SetSessionActiveChatRoomGroups#1_Response": steammessages_chat.CChatRoom_SetSessionActiveChatRoomGroups_Response,
    "ChatRoom.SetUserChatGroupPreferences#1_Request": steammessages_chat.CChatRoom_SetUserChatGroupPreferences_Request,
    "ChatRoom.SetUserChatGroupPreferences#1_Response": steammessages_chat.CChatRoom_SetUserChatGroupPreferences_Response,
    "ChatRoom.DeleteChatMessages#1_Request": steammessages_chat.CChatRoom_DeleteChatMessages_Request,
    "ChatRoom.DeleteChatMessages#1_Response": steammessages_chat.CChatRoom_DeleteChatMessages_Response,
    "ChatRoomClient.NotifyIncomingChatMessage#1": steammessages_chat.CChatRoom_IncomingChatMessage_Notification,
    "ChatRoomClient.NotifyChatMessageModified#1": steammessages_chat.CChatRoom_ChatMessageModified_Notification,
    "ChatRoomClient.NotifyMemberStateChange#1": steammessages_chat.CChatRoom_MemberStateChange_Notification,
    "ChatRoomClient.NotifyChatRoomHeaderStateChange#1": steammessages_chat.CChatRoom_ChatRoomHeaderState_Notification,
    "ChatRoomClient.NotifyChatRoomGroupRoomsChange#1": steammessages_chat.CChatRoom_ChatRoomGroupRoomsChange_Notification,
    "ChatRoomClient.NotifyShouldRejoinChatRoomVoiceChat#1": steammessages_chat.CChatRoom_NotifyShouldRejoinChatRoomVoiceChat_Notification,
    "ChatRoomClient.NotifyChatGroupUserStateChanged#1": steammessages_chat.ChatRoomClient_NotifyChatGroupUserStateChanged_Notification,
    "ChatRoomClient.NotifyAckChatMessageEcho#1": steammessages_chat.CChatRoom_AckChatMessage_Notification,
    "ClanChatRooms.GetClanChatRoomInfo#1_Request": steammessages_chat.CClanChatRooms_GetClanChatRoomInfo_Request,
    "ClanChatRooms.GetClanChatRoomInfo#1_Response": steammessages_chat.CClanChatRooms_GetClanChatRoomInfo_Response,
    "FriendMessages.GetActiveMessageSessions#1_Request": steammessages_friendmessages.CFriendsMessages_GetActiveMessageSessions_Request,
    "FriendMessages.GetActiveMessageSessions#1_Response": steammessages_friendmessages.CFriendsMessages_GetActiveMessageSessions_Response,
    "FriendMessages.SendMessage#1_Request": steammessages_friendmessages.CFriendMessages_SendMessage_Request,
    "FriendMessages.SendMessage#1_Response": steammessages_friendmessages.CFriendMessages_SendMessage_Response,
    "FriendMessages.AckMessage#1": steammessages_friendmessages.CFriendMessages_AckMessage_Notification,
    "FriendMessagesClient.IncomingMessage#1": steammessages_friendmessages.CFriendMessages_IncomingMessage_Notification,
    "FriendMessagesClient.NotifyAckMessageEcho#1": steammessages_friendmessages.CFriendMessages_AckMessage_Notification,
    "UserAccount.CreateFriendInviteToken#1_Request": steammessages_useraccount.CUserAccount_CreateFriendInviteToken_Request,
    "UserAccount.CreateFriendInviteToken#1_Response": steammessages_useraccount.CUserAccount_CreateFriendInviteToken_Response,
    "UserAccount.GetFriendInviteTokens#1_Request": steammessages_useraccount.CUserAccount_GetFriendInviteTokens_Request,
    "UserAccount.GetFriendInviteTokens#1_Response": steammessages_useraccount.CUserAccount_GetFriendInviteTokens_Response,
    "UserAccount.ViewFriendInviteToken#1_Request": steammessages_useraccount.CUserAccount_ViewFriendInviteToken_Request,
    "UserAccount.ViewFriendInviteToken#1_Response": steammessages_useraccount.CUserAccount_ViewFriendInviteToken_Response,
    "UserAccount.RedeemFriendInviteToken#1_Request": steammessages_useraccount.CUserAccount_RedeemFriendInviteToken_Request,
    "UserAccount.RedeemFriendInviteToken#1_Response": steammessages_useraccount.CUserAccount_RedeemFriendInviteToken_Response,
    "UserAccount.RevokeFriendInviteToken#1_Request": steammessages_useraccount.CUserAccount_RevokeFriendInviteToken_Request,
    "UserAccount.RevokeFriendInviteToken#1_Response": steammessages_useraccount.CUserAccount_RevokeFriendInviteToken_Response,
}



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from datetime import datetime
from typing import TYPE_CHECKING

from .abc import Message

if TYPE_CHECKING:
    from .abc import BaseUser
    from .channel import ClanChannel, DMChannel, GroupChannel
    from .protobufs.steammessages_chat import CChatRoom_IncomingChatMessage_Notification as GroupMessageNotification
    from .protobufs.steammessages_friendmessages import (
        CFriendMessages_IncomingMessage_Notification as UserMessageNotification,
    )


__all__ = (
    "UserMessage",
    "GroupMessage",
)


class UserMessage(Message):
    """Represents a message from a User."""

    def __init__(self, proto: "UserMessageNotification", channel: "DMChannel"):
        super().__init__(channel)
        self.author = channel.participant
        self.content = proto.message.replace("\[", "[").replace("\\\\", "\\")
        self.created_at = datetime.utcfromtimestamp(proto.rtime32_server_timestamp)
        self.clean_content = proto.message_no_bbcode or self.content

    def __repr__(self):
        attrs = ("author", "channel")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<UserMessage {' '.join(resolved)}>"


class _GroupMessage(Message):
    def __init__(self, proto: "GroupMessageNotification", channel, author: "BaseUser"):
        super().__init__(channel)
        self.author = author
        self.content = proto.message.replace("\[", "[").replace("\\\\", "\\")
        self.created_at = datetime.utcfromtimestamp(proto.timestamp)
        self.clean_content = proto.message_no_bbcode or self.content

    def __repr__(self):
        attrs = ("author", "channel")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<{self.__class__.__name__} {' '.join(resolved)}>"


class GroupMessage(_GroupMessage):
    """Represents a message in a Group."""

    def __init__(
        self, proto: "GroupMessageNotification", channel: "GroupChannel", author: "BaseUser",
    ):
        super().__init__(proto, channel, author)


class ClanMessage(_GroupMessage):
    """Represents a message in a Clan."""

    def __init__(
        self, proto: "GroupMessageNotification", channel: "ClanChannel", author: "BaseUser",
    ):
        super().__init__(proto, channel, author)



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

__all__ = (
    "TF2",
    "Game",
    "CSGO",
    "DOTA2",
    "STEAM",
    "CUSTOM_GAME",
)

APP_ID_MAX = 2 ** 32

MAPPING = {
    "Team Fortress 2": [440, 2],
    "DOTA 2": [570, 2],
    "Counter Strike Global-Offensive": [730, 2],
    "Steam": [753, 6],
    440: ["Team Fortress 2", 2],
    570: ["DOTA 2", 2],
    730: ["Counter Strike Global-Offensive", 2],
    753: ["Steam", 6],
}


class Game:
    """Represents a Steam game.

    .. note::

        This class can be defined by users using the above parameters, or
        it can be from an API call this is when :meth:`~steam.User.fetch_games`
        is called.

    Parameters
    ----------
    title: Optional[:class:`str`]
        The game's title.
    app_id: Optional[:class:`int`]
        The game's app_id.
    context_id: Optional[:class:`int`]
        The games's context ID by default 2.

    Attributes
    -----------
    title: Optional[:class:`str`]
        The game's title.
    app_id: :class:`int`
        The game's app_id.
    context_id: :class:`int`
        The context id of the game normally 2.
    total_play_time: Optional[:class:`int`]
        The total time the game has been played for.
        Only applies to a :class:`~steam.User`'s games from :meth:`~steam.User.games`.
    icon_url: Optional[:class:`str`]
        The icon url of the game.
        Only applies to a :class:`~steam.User`'s games from :meth:`~steam.User.games`.
    logo_url: Optional[:class:`str`]
        The logo url of the game.
        Only applies to a :class:`~steam.User`'s games from :meth:`~steam.User.games`.
    stats_visible: Optional[:class:`bool`]
        Whether the game has publicly visible stats.
        Only applies to a :class:`~steam.User`'s games from :meth:`~steam.User.games`.
    """

    __slots__ = (
        "app_id",
        "title",
        "context_id",
        "total_play_time",
        "icon_url",
        "logo_url",
        "stats_visible",
    )

    def __init__(self, app_id: int = None, title: str = None, *, context_id: int = 2):
        if app_id is not None and title is None:
            try:
                self.app_id = int(app_id)
            except ValueError as exc:
                raise ValueError(f"app_id expected to be a digit not {repr(type(app_id).__name__)}") from exc
            if self.app_id < 0:
                raise ValueError("app_id cannot be negative")
            mapping = MAPPING.get(app_id)
            if mapping is not None:
                self.title = mapping[0]
                self.context_id = mapping[1]
            else:
                self.title = title
                self.context_id = context_id

        elif app_id is None and title is not None:
            mapping = MAPPING.get(title)
            if mapping is not None:
                self.app_id = mapping[0]
                self.context_id = mapping[1]
            else:
                self.app_id = app_id
                self.context_id = context_id
            self.title = title

        else:
            app_id = int(app_id)
            mapping = MAPPING.get(app_id)
            if mapping is not None:
                self.title = mapping[0]
                self.app_id = app_id
                self.context_id = mapping[1]
            else:
                self.title = title
                self.app_id = app_id
                self.context_id = context_id

            mapping = MAPPING.get(title)
            if mapping is not None:
                self.title = title
                self.app_id = mapping[0]
                self.context_id = mapping[1]
            else:
                self.title = title
                self.app_id = app_id
                self.context_id = context_id

        self.total_play_time = None
        self.icon_url = None
        self.logo_url = None
        self.stats_visible = None

    @classmethod
    def _from_api(cls, data):
        game = cls(app_id=data.get("appid"), title=data.get("name"))
        game.total_play_time = data.get("playtime_forever", 0)
        game.icon_url = data.get("img_icon_url")
        game.logo_url = data.get("img_logo_url")
        game.stats_visible = data.get("has_community_visible_stats", False)
        return game

    def __repr__(self):
        attrs = ("title", "app_id", "context_id")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<Game {' '.join(resolved)}>"

    def __eq__(self, other):
        return isinstance(other, Game) and (self.app_id == other.app_id or self.title == other.title)

    def to_dict(self) -> dict:
        if not self.is_steam_game():
            return {"game_id": str(self.app_id), "game_extra_info": self.title}
        return {"game_id": str(self.app_id)}

    def is_steam_game(self) -> bool:
        return self.app_id <= APP_ID_MAX


TF2 = Game(title="Team Fortress 2", app_id=440)
DOTA2 = Game(title="DOTA 2", app_id=570)
CSGO = Game(title="Counter Strike Global-Offensive", app_id=730)
STEAM = Game(title="Steam", app_id=753, context_id=6)


def CUSTOM_GAME(title: str):
    """Create a custom game instance
    for :meth:`~steam.Client.change_presence`.

    Examples: ::
        await client.change_presence(game=steam.CUSTOM_GAME('my cool game'))

    Parameters
    ----------
    title: :class:`str`
        The name of the game to set your playing status to
    """
    return Game(title=title, app_id=15190414816125648896, context_id=0)



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2015-2020 Rapptz
Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

Heavily inspired by
https://github.com/Rapptz/discord.py/blob/master/discord/ext/commands/core.py
"""

import asyncio
import inspect
from typing import TYPE_CHECKING, Any, Awaitable, Callable, Dict, List, Type, Union

import steam

from ...errors import ClientException
from . import converters
from .cooldown import BucketType, Cooldown
from .errors import BadArgument, MissingRequiredArgument

if TYPE_CHECKING:
    from ...client import EventType
    from .cog import Cog
    from .context import Context

__all__ = (
    "Command",
    "command",
)

CommandFuncType = Callable[["Context"], Awaitable[None]]


def to_bool(argument: str):
    lowered = argument.lower()
    if lowered in ("yes", "y", "true", "t", "1", "enable", "on"):
        return True
    elif lowered in ("no", "n", "false", "f", "0", "disable", "off"):
        return False
    else:
        raise BadArgument(f'"{lowered}" is not a recognised boolean option')


class Command:
    def __init__(self, func: CommandFuncType, **kwargs):
        if not asyncio.iscoroutinefunction(func):
            raise TypeError("Callback must be a coroutine.")

        self.callback = func
        self.checks: List[Callable[..., Awaitable[bool]]] = []
        self._cooldowns: List[Cooldown] = []
        self.params = inspect.signature(func).parameters
        self.name = kwargs.get("name") or func.__name__
        if not isinstance(self.name, str):
            raise TypeError("Name of a command must be a string.")

        self.enabled = kwargs.get("enabled", True)

        help_doc = kwargs.get("help")
        if help_doc is not None:
            help_doc = inspect.cleandoc(help_doc)
        else:
            help_doc = inspect.getdoc(func)
            if isinstance(help_doc, bytes):
                help_doc = help_doc.decode("utf-8")

        self.help = help_doc

        self.brief = kwargs.get("brief")
        self.usage = kwargs.get("usage")
        self.aliases = kwargs.get("aliases", [])

        try:
            for alias in self.aliases:
                if not type(alias) is str:
                    raise TypeError
        except TypeError:
            raise TypeError("aliases of a command must be an iterable containing only strings.")

        self.description = inspect.cleandoc(kwargs.get("description", ""))
        self.hidden = kwargs.get("hidden", False)
        self.cog: "Cog" = kwargs.get("cog")

    def __call__(self, *args, **kwargs):
        """|coro|
        Calls the internal callback that the command holds.

        .. note::
            This bypasses all mechanisms -- including checks, converters,
            invoke hooks, cooldowns, etc. You must take care to pass
            the proper arguments and types to this function.
        """
        if self.cog is not None:
            return self.callback(self.cog, *args, **kwargs)
        else:
            return self.callback(*args, **kwargs)

    def error(self, func: "EventType") -> "EventType":
        """Register an event to handle a commands
        ``on_error`` function.

        Example: ::

            @bot.command()
            async def raise_an_error(ctx):
                raise Exception('oh no an error')


            @raise_an_error.error
            async def on_error(ctx, error):
                print(f'{ctx.command.name} raised an exception {error}')
        """
        if not asyncio.iscoroutinefunction(func):
            raise TypeError("callback must be a coroutine.")
        self.on_error = func
        return func

    async def _parse_arguments(self, ctx: "Context") -> None:
        args = [ctx] if self.cog is None else [self.cog, ctx]
        kwargs = {}

        shlex = ctx.shlex
        iterator = iter(self.params.items())

        if self.cog is not None:
            # we have 'self' as the first parameter so just advance
            # the iterator and resume parsing
            try:
                next(iterator)
            except StopIteration:
                raise ClientException(f'Callback for {self.name} command is missing "self" parameter.')

        # next we have the 'ctx' as the next parameter
        try:
            next(iterator)
        except StopIteration:
            raise ClientException(f'Callback for {self.name} command is missing "ctx" parameter.')
        for name, param in iterator:
            if param.kind == param.POSITIONAL_OR_KEYWORD:
                argument = shlex.get_token()
                if argument is None:
                    transformed = await self._get_default(ctx, param)
                else:
                    transformed = await self.transform(ctx, param, argument)
                args.append(transformed)
            elif param.kind == param.KEYWORD_ONLY:
                # kwarg only param denotes "consume rest" semantics
                arg = " ".join(shlex)
                if not arg:
                    kwargs[name] = await self._get_default(ctx, param)
                else:
                    kwargs[name] = await self.transform(ctx, param, arg)
                break
            elif param.kind == param.VAR_KEYWORD:
                # we have received **kwargs
                arguments = list(shlex)
                if not arguments:
                    kwargs[name] = await self._get_default(ctx, param)

                annotation = param.annotation
                if annotation is inspect.Parameter.empty or annotation is dict:
                    annotation = Dict[str, str]  # default to {str: str}

                key_converter = self._get_converter(annotation.__args__[0])
                value_converter = self._get_converter(annotation.__args__[1])

                kv_pairs = [arg.split("=") for arg in arguments]
                kwargs[name] = {
                    await self._convert(ctx, key_converter, key): await self._convert(ctx, value_converter, value)
                    for (key, value) in kv_pairs
                }
                break

            elif param.kind == param.VAR_POSITIONAL:
                # same as *args
                for arg in shlex:
                    transformed = await self.transform(ctx, param, arg)
                    args.append(transformed)
                break
        ctx.args = tuple(args)
        ctx.kwargs = kwargs

    def transform(self, ctx, param: inspect.Parameter, argument: str) -> Awaitable[Any]:
        param_type = param.annotation if param.annotation is not param.empty else str
        converter = self._get_converter(param_type)
        return self._convert(ctx, converter, argument)

    def _get_converter(self, param_type: type) -> Union[converters.Converter, type]:
        if param_type.__name__ in dir(steam):  # find a converter
            converter = getattr(converters, f"{param_type.__name__}Converter", None)
            if converter is None:
                raise NotImplementedError(f"{param_type.__name__} does not have an associated converter")
            return converter
        return param_type

    async def _convert(
        self, ctx: "Context", converter: Union[Type[converters.Converter], type], argument: str,
    ):
        if isinstance(converter, converters.Converter):
            try:
                return await converter.convert(ctx, argument)
            except Exception as exc:
                raise BadArgument(f"{argument} failed to convert to type {converter.__name__ or str}") from exc
        if issubclass(converter, converters.Converter):
            try:
                return await converter.convert(converter, ctx, argument)
            except Exception as exc:
                raise BadArgument(f"{argument} failed to convert to type {converter.__name__ or str}") from exc
        else:
            if converter is bool:
                return to_bool(argument)
            try:
                return converter(argument)
            except TypeError as exc:
                raise BadArgument(f"{argument} failed to convert to type {converter.__name__ or str}") from exc

    async def _get_default(self, ctx, param: inspect.Parameter):
        if param.default is param.empty:
            raise MissingRequiredArgument(param)
        if inspect.isclass(param.default):
            if isinstance(param.default, converters.Default):
                return await param.default.default(ctx)
            if issubclass(param.default, converters.Default):
                return await param.default.default(param.default, ctx)
        return param.default

    def _parse_cooldown(self, ctx: "Context"):
        for cooldown in self._cooldowns:
            bucket = cooldown.bucket.get_bucket(ctx)
            cooldown(bucket)


def command(name: str = None, cls: Type[Command] = None, **attrs) -> Callable[..., Command]:
    r"""Register a coroutine as a :class:`~commands.Command`.

    Parameters
    ----------
    name: :class:`str`
        The name of the command.
        Will default to ``func.__name__``.
    cls: Type[:class:`Command`]
        The class to construct the command from. Defaults to
        :class:`Command`.
    \*\*attrs:
        The attributes to pass to the command's ``__init__``.
    """
    if cls is None:
        cls = Command

    def decorator(func: CommandFuncType) -> Command:
        if isinstance(func, Command):
            raise TypeError("Callback is already a command.")
        return cls(func, name=name, **attrs)

    return decorator


def cooldown(rate: int, per: float, bucket: BucketType) -> Callable[..., None]:
    """Mark a :class:`Command`'s cooldown.

    Parameters
    ----------
    rate: :class:`int`
        The amount of times a command can be executed
        before being put on cooldown.
    per: :class:`float`
        The amount of time to wait between cooldowns.
    bucket:
        The :class:`.BucketType` that the cooldown applies
        to.
    """

    def decorator(func: "Command") -> None:
        func._cooldowns.append(Cooldown(rate, per, bucket))

    return decorator
